import pytest
from unittest import mock
from unittest.mock import AsyncMock, patch
from blebox_uniapi.box import Box
from blebox_uniapi import error
from blebox_uniapi.jfollow import follow

pytestmark = pytest.mark.asyncio


@pytest.fixture
def mock_session():
    return mock.MagicMock(host="172.1.2.3", port=80)


@pytest.fixture
def sample_data():
    return {
        "id": "abcd1234ef",
        "type": "airSensor",
        "deviceName": "foobar",
        "fv": "1.23",
        "hv": "4.56",
        "apiLevel": "20180403",
    }


@pytest.fixture
def config(sample_data):
    return Box._match_device_config(sample_data)


async def test_without_type(mock_session, sample_data, config):
    del sample_data["type"]

    with pytest.raises(error.UnsupportedBoxResponse, match="has no type"):
        Box(mock_session, sample_data, config, None)


async def test_with_unknown_type(mock_session, sample_data):
    sample_data["type"] = "unknownBox"

    with pytest.raises(error.UnsupportedBoxResponse, match="not a supported type"):
        Box._match_device_config(sample_data)


async def test_without_name(mock_session, sample_data, config):
    del sample_data["deviceName"]

    with pytest.raises(error.UnsupportedBoxResponse, match="has no name"):
        Box(mock_session, sample_data, config, None)


async def test_without_firmware_version(mock_session, sample_data, config):
    del sample_data["fv"]

    with pytest.raises(error.UnsupportedBoxResponse, match="has no firmware version"):
        Box(mock_session, sample_data, config, None)


async def test_without_hardware_version(mock_session, sample_data, config):
    del sample_data["hv"]

    with pytest.raises(error.UnsupportedBoxResponse, match="has no hardware version"):
        Box(mock_session, sample_data, config, None)


async def test_without_api_level(mock_session, sample_data, config):
    del sample_data["apiLevel"]

    with pytest.raises(error.UnsupportedBoxVersion, match=r"unsupported version"):
        Box._match_device_config(sample_data)


async def test_json_path_extraction(mock_session, sample_data, config):
    # note: follow is thin wrapper over jmespath so there's no real reason to test it.
    # However, this tests makes sure we understand the syntax and know how it works.
    # We can also use it as a canary for any unexpected change in syntax that may come
    # in newer version.

    # succesfull extraction
    assert follow(["foo"], "[0]") == "foo"
    assert follow([{"foo": "3", "value": 4}], "[?foo=='3'].value") == [4]

    # "dud" extraction
    assert follow([{"foo": "ab", "value": 4}], "[?foo=='bc'].value") == []
    assert follow([{"value": 4}], "[1].value") is None
    assert follow({"value": 4}, "foo") is None
    assert follow({"foo": [4]}, "[?bar==`0`].value") is None


async def test_missing_device_id(mock_session, sample_data, config):
    del sample_data["id"]
    with pytest.raises(error.UnsupportedBoxResponse, match="has no id"):
        Box(mock_session, sample_data, config, None)


# Add more test cases for other missing fields (type, name, versions, etc.)


async def test_invalid_init(mock_session, sample_data, config):
    with mock.patch(
        "blebox_uniapi.sensor.SensorFactory.many_from_config",
        spec_set=True,
        autospec=True,
    ) as mock_sensor:
        mock_sensor.side_effect = KeyError
        with pytest.raises(
            error.UnsupportedBoxResponse, match=r"Failed to initialize:"
        ):
            Box(mock_session, sample_data, config, None)


async def test_properties(mock_session, sample_data, config):
    box = Box(mock_session, sample_data, config, None)
    assert box.name == "foobar"
    assert box.last_data is None
    assert box.type == "airSensor"
    assert box.model == "airSensor"
    assert box.unique_id == "abcd1234ef"
    assert box.firmware_version == "1.23"
    assert box.hardware_version == "4.56"
    assert box.brand == "BleBox"
    assert box.api_version == 20180403
    assert box.address == "172.1.2.3:80"


async def test_field_validations(mock_session, sample_data, config):
    box = Box(mock_session, sample_data, config, None)

    with pytest.raises(
        error.BadFieldExceedsMax,
        match=r"foobar.field1 is 123 which exceeds max \(100\)",
    ):
        box.check_int_range(123, "field1", 100, 0)

    with pytest.raises(
        error.BadFieldLessThanMin,
        match=r"foobar.field1 is 123 which is less than minimum \(200\)",
    ):
        box.check_int_range(123, "field1", 300, 200)

    with pytest.raises(error.BadFieldMissing, match=r"foobar.field1 is missing"):
        box.check_int(None, "field1", 300, 200)

    with pytest.raises(
        error.BadFieldNotANumber, match=r"foobar.field1 is '123' which is not a number"
    ):
        box.check_int("123", "field1", 300, 200)

    with pytest.raises(error.BadFieldMissing, match=r"foobar.field1 is missing"):
        box.check_hex_str(None, "field1", 300, 200)

    with pytest.raises(
        error.BadFieldNotAString, match=r"foobar.field1 is 123 which is not a string"
    ):
        box.check_hex_str(123, "field1", 300, 200)

    with pytest.raises(error.BadFieldMissing, match=r"foobar.field1 is missing"):
        box.check_rgbw(None, "field1")

    with pytest.raises(
        error.BadFieldNotAString, match=r"foobar.field1 is 123 which is not a string"
    ):
        box.check_rgbw(123, "field1")

    with pytest.raises(
        error.BadFieldNotRGBW, match=r"foobar.field1 is 123 which is not a rgbw string"
    ):
        box.check_rgbw("123", "field1")


async def test_available_firmware_version_none(mock_session, sample_data, config):
    box = Box(mock_session, sample_data, config, None)
    assert box.available_firmware_version is None


async def test_available_firmware_version_set(mock_session, sample_data, config):
    sample_data["availableFv"] = "2.0"
    box = Box(mock_session, sample_data, config, None)
    assert box.available_firmware_version == "2.0"


async def test_async_ota_update_calls_session(mock_session, sample_data, config):
    box = Box(mock_session, sample_data, config, None)
    mock_session.async_api_get = AsyncMock(return_value=None)
    await box.async_ota_update()
    mock_session.async_api_get.assert_called_once_with("/api/ota/update")


async def test_async_ota_check_updates_firmware_versions(mock_session, sample_data, config):
    box = Box(mock_session, sample_data, config, None)
    mock_session.async_api_get_ota = AsyncMock(return_value=None)
    mock_session.async_api_get = AsyncMock(
        return_value={"availableFv": "2.0", "fv": "1.5"}
    )
    with patch("asyncio.sleep", new=AsyncMock()):
        await box.async_ota_check()
    assert box.available_firmware_version == "2.0"
    assert box.firmware_version == "1.5"


async def test_async_ota_check_unwraps_device_key(mock_session, sample_data, config):
    box = Box(mock_session, sample_data, config, None)
    mock_session.async_api_get_ota = AsyncMock(return_value=None)
    mock_session.async_api_get = AsyncMock(
        return_value={"device": {"availableFv": "3.0", "fv": "2.5"}}
    )
    with patch("asyncio.sleep", new=AsyncMock()):
        await box.async_ota_check()
    assert box.available_firmware_version == "3.0"


async def test_async_ota_check_raises_on_none_info_response(mock_session, sample_data, config):
    box = Box(mock_session, sample_data, config, None)
    mock_session.async_api_get_ota = AsyncMock(return_value=None)
    mock_session.async_api_get = AsyncMock(return_value=None)
    with patch("asyncio.sleep", new=AsyncMock()):
        with pytest.raises(error.UnsupportedBoxResponse):
            await box.async_ota_check()


async def test_async_ota_check_returns_silently_when_no_available_fv(mock_session, sample_data, config):
    box = Box(mock_session, sample_data, config, None)
    mock_session.async_api_get_ota = AsyncMock(return_value=None)
    mock_session.async_api_get = AsyncMock(return_value={"fv": "1.0"})
    with patch("asyncio.sleep", new=AsyncMock()):
        await box.async_ota_check()
    assert box.available_firmware_version is None



from unittest.mock import Mock
import pytest

from blebox_uniapi.button import Button
from blebox_uniapi.box import Box
from blebox_uniapi.box_types import BOX_TYPE_CONF


@pytest.fixture
def product():
    return Mock(spec=Box)


@pytest.fixture
def tv_lift_box_0(product):
    product.type = "tvLiftBox"
    extended_state = {"tvLift": {"controlType": 4}}
    many = Button.many_from_config(
        product,
        BOX_TYPE_CONF["tvLiftBox"][20200518]["buttons"],
        extended_state=extended_state,
    )
    assert len(many) == 3
    return many[0]


@pytest.fixture
def tv_lift_box_1(product):
    product.type = "tvLiftBox"
    extended_state = {"tvLift": {"controlType": 4}}
    many = Button.many_from_config(
        product,
        BOX_TYPE_CONF["tvLiftBox"][20200518]["buttons"],
        extended_state=extended_state,
    )
    assert len(many) == 3
    return many[1]


async def test_tv_lift_0_box_pressed(tv_lift_box_0: Button, product: Box):
    await tv_lift_box_0.set()
    product.async_api_command.assert_called_with("set", "open_or_stop")
    assert tv_lift_box_0.control_type


async def test_tv_lift_1_box_pressed(tv_lift_box_1: Button, product: Box):
    await tv_lift_box_1.set()
    product.async_api_command.assert_called_with("set", "close_or_stop")
    assert tv_lift_box_1.control_type



"""PyTest fixtures and test helpers."""

import copy
import json as _json
import re
from datetime import date, timedelta

import logging

from unittest.mock import AsyncMock
from deepmerge import Merger

from unittest import mock

from aiohttp import ClientResponseError, ClientSession

import pytest

from blebox_uniapi.box_types import get_latest_conf
from blebox_uniapi.session import ApiHost
from blebox_uniapi.box import Box
from blebox_uniapi.error import UnsupportedBoxVersion

_LOGGER = logging.getLogger(__name__)

retype = type(re.compile(""))


@pytest.fixture
def aioclient_mock():
    return AsyncMock(spec=ClientSession)


def array_merge(config, path, base, nxt):
    """Replace an array element with the merge result of elements."""
    if len(nxt):
        if isinstance(nxt[0], dict):
            index = 0
            for item in nxt:
                if not isinstance(item, dict):
                    raise NotImplementedError
                my_merger.merge(base[index], item)
                index += 1
            return base
        elif isinstance(nxt[0], int):
            return [nxt[0]]
        else:
            raise NotImplementedError


my_merger = Merger(
    # pass in a list of tuple, with the
    # strategies you are looking to apply
    # to each type.
    [(list, [array_merge]), (dict, ["merge"])],
    # next, choose the fallback strategies,
    # applied to all other types:
    ["override"],
    # finally, choose the strategies in
    # the case where the types conflict:
    ["override"],
)


def jmerge(base, ext):
    """Create new fixtures by adjusting existing ones."""

    result = copy.deepcopy(base)
    my_merger.merge(result, _json.loads(ext))
    return result


def future_date(delta_days=300):
    """Generate future date string in 'YYYYMMDD' format."""
    future_date = date.today() + timedelta(days=delta_days)
    return future_date.strftime("%Y%m%d")


HTTP_MOCKS = {}


def json_get_expect(mock, url, **kwargs):
    json = kwargs["json"]

    if mock not in HTTP_MOCKS:
        HTTP_MOCKS[mock] = {}
    HTTP_MOCKS[mock][url] = json

    class EffectWhenGet:
        def __init__(self, key):
            self._key = key

        def __call__(self, url, **kwargs):
            data = HTTP_MOCKS[self._key][url]
            response = _json.dumps(data).encode("utf-8")
            status = 200
            return AiohttpClientMockResponse(
                "GET", url, status, response, headers={"content-type": "application/json"}
            )

    mock.get = AsyncMock(side_effect=EffectWhenGet(mock))


def json_post_expect(mock, url, **kwargs):
    json = kwargs["json"]
    params = kwargs["params"]

    # TODO: check
    # headers = kwargs.get("headers")
    if mock not in HTTP_MOCKS:
        HTTP_MOCKS[mock] = {}
    if url not in HTTP_MOCKS[mock]:
        HTTP_MOCKS[mock][url] = {}

    HTTP_MOCKS[mock][url][params] = json

    class EffectWhenPost:
        def __init__(self, key):
            self._key = key

        def __call__(self, url, **kwargs):
            # TODO: timeout
            params = kwargs.get("data")
            # TODO: better checking of params (content vs raw json)
            data = HTTP_MOCKS[self._key][url][params]
            response = _json.dumps(data).encode("utf-8")
            status = 200
            return AiohttpClientMockResponse(
                "POST", url, status, response, headers={"content-type": "application/json"}
            )

    mock.post = AsyncMock(side_effect=EffectWhenPost(mock))


class DefaultBoxTest:
    """Base class with methods common to BleBox integration tests."""

    IP = "172.0.0.1"
    LOGGER = _LOGGER

    async def async_entities(self, session):
        """Get a created entity at the given index."""

        host = self.IP
        port = 80
        timeout = 2
        api_host = ApiHost(host, port, timeout, session, None, self.LOGGER)
        product = await Box.async_from_host(api_host)
        return [
            self.ENTITY_CLASS(feature) for feature in product.features[self.DEVCLASS]
        ]

    async def allow_get_info(self, aioclient_mock, info=None):
        """Stub a HTTP GET request for the device state."""

        data = self.DEVICE_INFO if info is None else info

        json_get_expect(
            aioclient_mock, f"http://{self.IP}:80/api/device/state", json=data
        )
        if (hasattr(self, "DEVICE_EXTENDED_INFO")) and (
            path := getattr(self, "DEVICE_EXTENDED_INFO_PATH")
        ):
            data = self.DEVICE_EXTENDED_INFO or info
            json_get_expect(
                aioclient_mock,
                f"http://{self.IP}:80/{path.lstrip('/')}",
                json=data,
            )

    def allow_get_state(self, aioclient_mock, data):
        """Stub a HTTP GET request for the product-specific state."""
        json_get_expect(
            aioclient_mock, f"http://{self.IP}:80/{self.DEV_INFO_PATH}", json=data
        )

    def allow_get(self, aioclient_mock, api_path, data):
        """Stub a HTTP GET request."""
        json_get_expect(
            aioclient_mock, f"http://{self.IP}:80/{api_path[1:]}", json=data
        )

    async def allow_post(self, code, aioclient_mock, api_path, post_data, response):
        """Stub a HTTP POST request."""
        json_post_expect(
            aioclient_mock,
            f"http://{self.IP}:80/{api_path[1:]}",
            params=post_data,
            headers={"content-type": "application/json"},
            json=response,
        )
        await code()

    # TODO: rename?
    async def updated(self, aioclient_mock, state, index=0):
        """Return an entry on which update has already been called."""

        await self.allow_get_info(aioclient_mock)
        entity = (await self.async_entities(aioclient_mock))[index]

        self.allow_get_state(aioclient_mock, state)
        await entity.async_update()
        return entity

    async def test_future_version(self, aioclient_mock):
        """
        Test support for future versions, that is last supported entry in config type file.
        """
        await self.allow_get_info(aioclient_mock, self.DEVICE_INFO_FUTURE)
        entity = (await self.async_entities(aioclient_mock))[0]

        assert entity._feature.product._config is get_latest_conf(
            entity._feature.product.type
        )

    async def test_latest_version(self, aioclient_mock):
        """
        Test support for latest versions, that is last supported entry in config type file.
        """
        await self.allow_get_info(aioclient_mock, self.DEVICE_INFO_LATEST)
        entity = (await self.async_entities(aioclient_mock))[0]

        assert entity._feature.product._config is get_latest_conf(
            entity._feature.product.type
        )

    async def test_unsupported_version(self, aioclient_mock):
        """Test version support."""

        # only gateBox is same
        if self.DEVICE_INFO != self.DEVICE_INFO_UNSUPPORTED:
            await self.allow_get_info(aioclient_mock, self.DEVICE_INFO_UNSUPPORTED)
            with pytest.raises(UnsupportedBoxVersion):
                await self.async_entities(aioclient_mock)

    async def test_unspecified_version(self, aioclient_mock):
        """
        Test default_api_level when api level is not specified in device info.

        """
        if self.DEVICE_INFO_UNSPECIFIED_API is not None:
            await self.allow_get_info(aioclient_mock, self.DEVICE_INFO_UNSPECIFIED_API)
            with pytest.raises(UnsupportedBoxVersion):
                await self.async_entities(aioclient_mock)


class AiohttpClientMockResponse:
    """Mock Aiohttp client response."""

    def __init__(
        self, method, url, status, response, cookies=None, exc=None, headers=None
    ):
        """Initialize a fake response."""
        self.method = method
        self._url = url
        self.status = status
        self.response = response
        self.exc = exc

        self._headers = headers or {}
        self._cookies = {}

        if cookies:
            for name, data in cookies.items():
                cookie = mock.MagicMock()
                cookie.value = data
                self._cookies[name] = cookie

    @property
    def headers(self):
        return self._headers

    @property
    def cookies(self):
        return self._cookies

    @property
    def url(self):
        return self._url

    @property
    def content_type(self):
        return self._headers.get("content-type")

    async def read(self):
        return self.response

    async def text(self, encoding="utf-8"):
        return self.response.decode(encoding)

    async def json(self, encoding="utf-8"):
        return _json.loads(self.response.decode(encoding))

    async def release(self):
        pass

    def raise_for_status(self):
        """Raise error if status is 400 or higher."""
        if self.status >= 400:
            request_info = mock.Mock(real_url="http://example.com")
            # TODO: coverage
            raise ClientResponseError(
                request_info=request_info,
                history=None,
                code=self.status,
                headers=self.headers,
            )

    def close(self):
        pass


class CommonEntity:
    def __init__(self, feature):
        self._feature = feature

    @property
    def name(self):
        return self._feature.full_name

    @property
    def unique_id(self):
        return self._feature.unique_id

    async def async_update(self):
        await self._feature.async_update()

    @property
    def device_info(self):
        product = self._feature.product
        return {
            "name": product.name,
            "mac": product.unique_id,
            "manufacturer": product.brand,
            "model": product.model,
            "sw_version": product.firmware_version,
        }



from typing import Union, Optional, TYPE_CHECKING

from .feature import Feature

if TYPE_CHECKING:
    from .box import Box


class BinarySensor(Feature):
    """Class representing sensor with bool state."""

    def __init__(
        self,
        product: "Box",
        alias: str,
        methods: dict,
        sensor_id: Optional[int] = None,
        name: Optional[str] = None,
    ):
        self._sensor_id = sensor_id
        self._name = name
        super().__init__(product, alias, methods)

    @property
    def index(self) -> Optional[int]:
        return self._sensor_id

    @property
    def name(self) -> Optional[str]:
        return self._name

    @classmethod
    def many_from_config(
        cls, product, box_type_config, extended_state
    ) -> list["Feature"]:
        type_class_map = {
            "rain": Rain,
            "flood": Flood,
            "openStatus": Open,
            "binary": Input,
        }

        output_list = list()
        sensors_list = extended_state.get("multiSensor").get("sensors", {})
        alias, methods = box_type_config[0]

        for sensor in sensors_list:
            sensor_type = sensor.get("type")
            sensor_id = sensor.get("id")

            if sensor_type in type_class_map:
                klass = type_class_map[sensor_type]

                if methods.get(sensor_type) is not None:
                    value_method = {sensor_type: methods[sensor_type](sensor_id)}
                    output_list.append(
                        klass(
                            product=product,
                            alias=sensor_type + "_" + str(sensor_id),
                            methods=value_method,
                            sensor_id=sensor_id,
                            name=sensor.get("name"),
                        )
                    )

        return output_list


class Rain(BinarySensor):
    def __init__(
        self,
        product: "Box",
        alias: str,
        methods: dict,
        sensor_id: Optional[int] = None,
        name: Optional[str] = None,
    ):
        self._device_class = "moisture"
        super().__init__(product, alias, methods, sensor_id=sensor_id, name=name)

    @property
    def state(self) -> bool:
        return self._current > 0

    @property
    def device_class(self) -> str:
        return self._device_class

    def _read_rain(self, field: str) -> Union[float, int, None]:
        product = self._product
        if product.last_data is not None:
            raw = self.raw_value(field)
            if raw is not None:  # no reading
                return self.raw_value("rain")
        return 0

    def after_update(self) -> None:
        self._current = self._read_rain("rain")


class Flood(BinarySensor):
    def __init__(
        self,
        product: "Box",
        alias: str,
        methods: dict,
        sensor_id: Optional[int] = None,
        name: Optional[str] = None,
    ):
        self._device_class = "moisture"
        super().__init__(product, alias, methods, sensor_id=sensor_id, name=name)

    @property
    def state(self) -> bool:
        return self._current > 0

    @property
    def device_class(self) -> str:
        return self._device_class

    def _read_flood(self, field: str) -> Union[float, int, None]:
        product = self._product

        if product.last_data is not None:
            raw = self.raw_value(field)
            if raw is not None:  # no reading
                return self.raw_value("flood")
        return 0

    def after_update(self) -> None:
        self._current = self._read_flood("flood")


class Open(BinarySensor):
    def __init__(
        self,
        product: "Box",
        alias: str,
        methods: dict,
        sensor_id: Optional[int] = None,
        name: Optional[str] = None,
    ):
        self._device_class = "open"
        super().__init__(product, alias, methods, sensor_id=sensor_id, name=name)

    @property
    def state(self) -> bool:
        return self._current != 4  # 4 = closed, 0-3 = open

    def _read_input(self, field: str) -> Union[float, int, None]:
        product = self._product

        if product.last_data is not None:
            raw = self.raw_value(field)
            if raw is not None:  # no reading
                return self.raw_value("openStatus")
        return 0

    def after_update(self) -> None:
        self._current = self._read_input("openStatus")


class Input(BinarySensor):
    def __init__(
        self,
        product: "Box",
        alias: str,
        methods: dict,
        sensor_id: Optional[int] = None,
        name: Optional[str] = None,
    ):
        self._device_class = "input"
        super().__init__(product, alias, methods, sensor_id=sensor_id, name=name)

    @property
    def state(self) -> bool:
        return self._current > 0

    def _read_input(self, field: str) -> Union[float, int, None]:
        product = self._product

        if product.last_data is not None:
            raw = self.raw_value(field)
            if raw is not None:  # no reading
                return self.raw_value("binary")
        return 0

    def after_update(self) -> None:
        self._current = self._read_input("binary")



import pytest

from blebox_uniapi.session import ApiHost
from blebox_uniapi.box import Box
from blebox_uniapi import error

from .conftest import json_get_expect


@pytest.fixture
def data():
    return {
        "id": "abcd1234ef",
        "type": "unknownBox",
        "deviceName": "foobar",
        "fv": "1.23",
        "hv": "4.56",
        "apiLevel": "20180403",
    }


class TestUnknownDevice:
    async def test_unknown_product(self, aioclient_mock, data):
        host = "172.1.2.3"
        with pytest.raises(error.UnsupportedBoxResponse, match=r"unknownBox"):
            full_data = {"device": data}
            json_get_expect(
                aioclient_mock, f"http://{host}:80/api/device/state", json=full_data
            )

            port = 80
            timeout = 2
            api_host = ApiHost(host, port, timeout, aioclient_mock, None, None)
            await Box.async_from_host(api_host)

    async def test_unknown_product_without_device_section(self, aioclient_mock, data):
        host = "172.1.2.3"
        with pytest.raises(error.UnsupportedBoxResponse, match=r"unknownBox"):
            json_get_expect(
                aioclient_mock, f"http://{host}:80/api/device/state", json=data
            )

            port = 80
            timeout = 2
            api_host = ApiHost(host, port, timeout, aioclient_mock, None, None)
            await Box.async_from_host(api_host)

    async def test_unknown_product_without_device_and_type(self, aioclient_mock, data):
        host = "172.1.2.3"
        with pytest.raises(error.UnsupportedBoxResponse, match=r"has no type"):
            del data["type"]
            json_get_expect(
                aioclient_mock, f"http://{host}:80/api/device/state", json=data
            )

            port = 80
            timeout = 2
            api_host = ApiHost(host, port, timeout, aioclient_mock, None, None)
            await Box.async_from_host(api_host)



"""BleBox light entities tests."""

import json

from .conftest import CommonEntity, DefaultBoxTest, future_date, jmerge

from blebox_uniapi.box_types import get_latest_api_level
from blebox_uniapi.light import Light

# TODO: remove
import colorsys

import pytest

# TODO: remove
ATTR_BRIGHTNESS = "brightness"
ATTR_HS_COLOR = "ATTR_HS_COLOR"
ATTR_WHITE_VALUE = "ATTR_WHITE_VALUE"
ATTR_RGB_COLOR = "rgb_color"
ATTR_RGBW_COLOR = "rgbw_color"
ATTR_EFFECT = "effect"
ATTR_COLOR_TEMP = "color_temp"
ATTR_RGBWW_COLOR = "rgbww_color"
SUPPORT_BRIGHTNESS = 1
SUPPORT_COLOR = 2
SUPPORT_WHITE_VALUE = 4
COLOR_MODE_BRIGHTNESS = "brightness"
COLOR_MODE_COLOR_TEMP = "color_temp"
COLOR_MODE_ONOFF = "onoff"
COLOR_MODE_RGB = "rgb"
COLOR_MODE_RGBW = "rgbw"
COLOR_MODE_RGBWW = "rgbww"

COLOR_MODE_MAP = {
    1: COLOR_MODE_RGBW,
    2: COLOR_MODE_RGB,
    3: COLOR_MODE_BRIGHTNESS,
    4: COLOR_MODE_RGBW,
    # RGB and Brightness 2 and 3 implementation difference, if W hex is not null only this or RGB + Brightness separated with mask
    5: COLOR_MODE_COLOR_TEMP,
    6: COLOR_MODE_COLOR_TEMP,  # two instances
    7: COLOR_MODE_RGBWW,
}

# NOTE: copied from Home Assistant color util module


def rgb_hex_to_rgb_list(hex_string: str):
    """Return an RGB color value list from a hex color string."""
    return [
        int(hex_string[i : i + len(hex_string) // 3], 16)
        for i in range(0, len(hex_string), len(hex_string) // 3)
    ]


def color_hsv_to_RGB(iH: float, iS: float, iV: float):
    """Convert an hsv color into its rgb representation.

    Hue is scaled 0-360
    Sat is scaled 0-100
    Val is scaled 0-100
    """
    fRGB = colorsys.hsv_to_rgb(iH / 360, iS / 100, iV / 100)
    return (int(fRGB[0] * 255), int(fRGB[1] * 255), int(fRGB[2] * 255))


def color_hs_to_RGB(iH: float, iS: float):
    """Convert an hsv color into its rgb representation."""
    return color_hsv_to_RGB(iH, iS, 100)


def color_RGB_to_hs(iR: float, iG: float, iB: float):
    """Convert an rgb color to its hs representation."""
    return color_RGB_to_hsv(iR, iG, iB)[:2]


def color_RGB_to_hsv(iR: float, iG: float, iB: float):
    """Convert an rgb color to its hsv representation.

    Hue is scaled 0-360
    Sat is scaled 0-100
    Val is scaled 0-100
    """
    fHSV = colorsys.rgb_to_hsv(iR / 255.0, iG / 255.0, iB / 255.0)
    return round(fHSV[0] * 360, 3), round(fHSV[1] * 100, 3), round(fHSV[2] * 100, 3)


def color_rgb_to_hex(r: int, g: int, b: int) -> str:
    """Return a RGB color from a hex color string."""
    return "{0:02x}{1:02x}{2:02x}".format(round(r), round(g), round(b))


class BleBoxLightEntity(CommonEntity):
    """Representation of BleBox lights."""

    @property
    def is_on(self) -> bool:
        """Return if light is on."""
        return self._feature.is_on

    @property
    def brightness(self):
        """Return the brightness."""
        return self._feature.brightness

    @property
    def color_temp(self):
        """Return color temperature."""
        return self._feature.color_temp

    @property
    def color_mode(self):
        """Return the color mode. Set values to _attr_ibutes if needed."""
        color_mode_tmp = COLOR_MODE_MAP.get(self._feature.color_mode, COLOR_MODE_ONOFF)
        if color_mode_tmp == COLOR_MODE_COLOR_TEMP:
            self._attr_min_mireds = 1
            self._attr_max_mireds = 255

        return color_mode_tmp

    @property
    def effect_list(self):
        """Return the list of supported effects."""
        return self._feature.effect_list

    @property
    def effect(self):
        """Return the current effect."""
        return self._feature.effect

    @property
    def rgb_color(self):
        """Return value for rgb."""
        if (rgb_hex := self._feature.rgb_hex) is None:
            return None
        return tuple(
            self._feature.normalise_elements_of_rgb(
                self._feature.rgb_hex_to_rgb_list(rgb_hex)[0:3]
            )
        )

    @property
    def rgbw_color(self):
        """Return the hue and saturation."""
        if (rgbw_hex := self._feature.rgbw_hex) is None:
            return None
        return tuple(self._feature.rgb_hex_to_rgb_list(rgbw_hex)[0:4])

    @property
    def rgbww_color(self):
        """Return value for rgbww."""
        if (rgbww_hex := self._feature.rgbww_hex) is None:
            return None
        return tuple(self._feature.rgb_hex_to_rgb_list(rgbww_hex))

    async def async_turn_on(self, **kwargs):
        """Turn the light on."""
        rgbw = kwargs.get(ATTR_RGBW_COLOR)
        brightness = kwargs.get(ATTR_BRIGHTNESS)
        effect = kwargs.get(ATTR_EFFECT)
        color_temp = kwargs.get(ATTR_COLOR_TEMP)
        rgbww = kwargs.get(ATTR_RGBWW_COLOR)
        feature = self._feature
        value = feature.sensible_on_value
        rgb = kwargs.get(ATTR_RGB_COLOR)

        if rgbw is not None:
            value = list(rgbw)
        if color_temp is not None:
            value = feature.return_color_temp_with_brightness(
                int(color_temp), self.brightness
            )

        if rgbww is not None:
            value = list(rgbww)

        if rgb is not None:
            if self.color_mode == COLOR_MODE_RGB and brightness is None:
                brightness = self.brightness
            value = list(rgb)

        if brightness is not None:
            if self.color_mode == ATTR_COLOR_TEMP:
                value = feature.return_color_temp_with_brightness(
                    self.color_temp, brightness
                )
            else:
                value = feature.apply_brightness(value, brightness)

        if effect is not None:
            effect_value = self.effect_list.index(effect)
            await self._feature.async_api_command("effect", effect_value)
        else:
            await self._feature.async_on(value)

    async def async_turn_off(self, **kwargs):
        """Turn the light off."""
        await self._feature.async_off()


class TestDimmer(DefaultBoxTest):
    """Tests for BleBox dimmerBox."""

    DEVCLASS = "lights"
    ENTITY_CLASS = BleBoxLightEntity

    DEV_INFO_PATH = "api/dimmer/state"
    DEVICE_INFO = json.loads(
        """
    {
        "device": {
            "deviceName": "My dimmer",
            "type": "dimmerBox",
            "fv": "0.247",
            "hv": "0.2",
            "id": "1afe34e750b8",
            "apiLevel": "20170829"
        },
        "network": {
            "ip": "192.168.1.239",
            "ssid": "myWiFiNetwork",
            "station_status": 5,
            "apSSID": "dimmerBox-ap",
            "apPasswd": ""
        },
        "dimmer": {
            "loadType": 7,
            "currentBrightness": 65,
            "desiredBrightness": 65,
            "temperature": 39,
            "overloaded": false,
            "overheated": false
        }
    }
    """
    )

    def patch_version(apiLevel):
        """Generate a patch for a JSON state fixture."""
        return f"""
        {{ "device": {{ "apiLevel": {apiLevel} }} }}
        """

    DEVICE_INFO_FUTURE = jmerge(DEVICE_INFO, patch_version(future_date()))
    DEVICE_INFO_LATEST = jmerge(
        DEVICE_INFO, patch_version(get_latest_api_level("dimmerBox"))
    )
    DEVICE_INFO_UNSUPPORTED = jmerge(
        DEVICE_INFO, patch_version(20140828)
    )  # fake version, default_api_level implemented

    DEVICE_INFO_UNSPECIFIED_API = json.loads(
        """
    {
        "device": {
            "deviceName": "My dimmer",
            "type": "dimmerBox",
            "fv": "0.247",
            "hv": "0.2",
            "id": "1afe34e750b8"
        },
        "network": {
            "ip": "192.168.1.239",
            "ssid": "myWiFiNetwork",
            "station_status": 5,
            "apSSID": "dimmerBox-ap",
            "apPasswd": ""
        },
        "dimmer": {
            "loadType": 7,
            "currentBrightness": 65,
            "desiredBrightness": 65,
            "temperature": 39,
            "overloaded": false,
            "overheated": false
        }
    }
    """
    )

    STATE_DEFAULT = json.loads(
        """
    {
        "dimmer": {
            "loadType": 7,
            "currentBrightness": 11,
            "desiredBrightness": 53,
            "temperature": 29,
            "overloaded": false,
            "overheated": false
        }
    }
    """
    )

    def patch_state(current, desired):
        """Generate a patch for a JSON state fixture."""
        return f"""
        {{
            "dimmer": {{
                "currentBrightness": {current},
                "desiredBrightness": {desired}
            }}
        }}
        """

    STATE_OFF = jmerge(STATE_DEFAULT, patch_state(0, 0))
    STATE_ON_DEFAULT = jmerge(STATE_DEFAULT, patch_state(238, 255))
    STATE_ON_BRIGHT = jmerge(STATE_DEFAULT, patch_state(201, 202))

    async def test_init(self, aioclient_mock):
        """Test cover default state."""
        await self.allow_get_info(aioclient_mock)
        entity = (await self.async_entities(aioclient_mock))[0]

        assert entity.name == "My dimmer (dimmerBox#brightness)"
        assert entity.unique_id == "BleBox-dimmerBox-1afe34e750b8-brightness"

        # assert entity.supported_features & SUPPORT_BRIGHTNESS
        assert entity.brightness is None

        assert entity.is_on is None

    async def test_device_info(self, aioclient_mock):
        await self.allow_get_info(aioclient_mock, self.DEVICE_INFO)
        entity = (await self.async_entities(aioclient_mock))[0]
        assert entity.device_info["name"] == "My dimmer"
        assert entity.device_info["mac"] == "1afe34e750b8"
        assert entity.device_info["manufacturer"] == "BleBox"
        assert entity.device_info["model"] == "dimmerBox"
        assert entity.device_info["sw_version"] == "0.247"

    async def test_update(self, aioclient_mock):
        """Test light updating."""

        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT)

        assert entity.brightness == 53
        assert entity.is_on is True

    async def allow_set_brightness(self, code, aioclient_mock, value, response):
        """Set up mock for HTTP POST simulating brightness change."""
        await self.allow_post(
            code,
            aioclient_mock,
            "/api/dimmer/set",
            '{"dimmer":{"desiredBrightness": ' + str(value) + "}}",
            response,
        )

    async def test_on(self, aioclient_mock):
        """Test light on."""
        entity = await self.updated(aioclient_mock, self.STATE_OFF)
        assert entity.is_on is False

        async def turn_on():
            await entity.async_turn_on()

        await self.allow_set_brightness(
            turn_on, aioclient_mock, 255, self.STATE_ON_DEFAULT
        )

        assert entity.is_on is True
        # TODO: is max brightness a good default?
        assert entity.brightness == 255

    async def test_on_with_brightness(self, aioclient_mock):
        """Test light on with a brightness value."""
        entity = await self.updated(aioclient_mock, self.STATE_OFF)
        assert entity.is_on is False

        async def turn_on():
            await entity.async_turn_on(brightness=202)

        await self.allow_set_brightness(
            turn_on, aioclient_mock, 202, self.STATE_ON_BRIGHT
        )

        assert entity.is_on is True
        assert entity.brightness == 202  # as if desired brightness not reached yet

    async def test_off(self, aioclient_mock):
        """Test light off."""
        entity = await self.updated(aioclient_mock, self.STATE_ON_DEFAULT)
        assert entity.is_on is True

        async def turn_off():
            await entity.async_turn_off()

        await self.allow_set_brightness(turn_off, aioclient_mock, 0, self.STATE_OFF)

        assert entity.is_on is False
        assert entity.brightness == 0

    async def test_unspecified_version(self, aioclient_mock):
        """
        As default_api_level implemented for this class of devices, test method shall be omitted.
        """
        pass


class TestWLightBoxS(DefaultBoxTest):
    """Tests for BleBox wLightBoxS."""

    DEVCLASS = "lights"
    ENTITY_CLASS = BleBoxLightEntity

    DEVICE_EXTENDED_INFO_PATH = "/api/rgbw/extended/state"
    DEV_INFO_PATH = "api/rgbw/state"
    DEVICE_INFO = json.loads(
        """
    {
        "device": {
            "deviceName": "My wLightBoxS",
            "type": "wLightBox",
            "product":"wLightBoxS",
            "fv": "0.924",
            "hv": "0.1",
            "universe": 0,
            "id": "1afe34e750b8",
            "apiLevel": 20200229,
            "ip": "192.168.9.13",
            "availableFv": null
        }
    }
    """
    )

    DEVICE_INFO2 = {
        "device": {
            "deviceName": "My wLightBoxS",
            "type": "wLightBox",
            "product": "wLightBoxS",
            "hv": "s_0.1",
            "fv": "0.1022",
            "universe": 0,
            "apiLevel": "20200229",
            "id": "ce50e32d2707",
            "ip": "192.168.1.25",
            "availableFv": None,
        }
    }

    DEVICE_EXTENDED_INFO = {
        "rgbw": {
            "desiredColor": "f5",
            "currentColor": "f5",  # 245 in decimal
            "lastOnColor": "f5",
            "durationsMs": {"colorFade": 1000, "effectFade": 1000, "effectStep": 1000},
            "effectID": 0,
            "colorMode": 3,
            "favColors": {"0": "ff", "1": "00", "2": "c0", "3": "40", "4": "00"},
            "effectsNames": {"0": "NONE", "1": "FADE", "2": "Stroboskop", "3": "BELL"},
        }
    }

    def patch_version(apiLevel):
        """Generate a patch for a JSON state fixture."""
        return f"""
        {{ "device": {{ "apiLevel": {apiLevel} }} }}
        """

    DEVICE_INFO_FUTURE = jmerge(DEVICE_INFO, patch_version(future_date()))
    DEVICE_INFO_LATEST = jmerge(
        DEVICE_INFO, patch_version(get_latest_api_level("wLightBoxS"))
    )
    DEVICE_INFO_UNSUPPORTED = jmerge(DEVICE_INFO, patch_version(20140717))

    DEVICE_INFO_UNSPECIFIED_API = json.loads(
        """
    {
        "device": {
            "deviceName": "My wLightBoxS",
            "type": "wLightBoxS",
            "fv": "0.247",
            "hv": "0.2",
            "id": "1afe34e750b8"
        },
        "network": {
            "ip": "192.168.1.239",
            "ssid": "myWiFiNetwork",
            "station_status": 5,
            "apSSID": "wLightBoxS-ap",
            "apPasswd": ""
        },
        "rgbw": {
            "desiredColor": "e3",
            "currentColor": "df",
            "fadeSpeed": 255,
            "effectID": 0
        }
    }
    """
    )

    STATE_ON = json.loads(
        """
        {
            "rgbw": {
                "desiredColor": "ab",
                "currentColor": "cd",
                "fadeSpeed": 255,
                "effectID": 0,
                "colorMode": 3
            }
        }
    """
    )

    STATE_FULL_ON = json.loads(
        """
        {
            "rgbw": {
                "desiredColor": "ff",
                "currentColor": "ce",
                "fadeSpeed": 255,
                "effectID": 0
            }
        }
    """
    )

    STATE_OFF = json.loads(
        """
        {
            "rgbw": {
                "desiredColor": "00",
                "currentColor": "00",
                "fadeSpeed": 255,
                "effectID": 0
            }
        }
    """
    )

    STATE_DEFAULT = STATE_ON

    async def test_init(self, aioclient_mock):
        """Test cover default state."""
        await self.allow_get_info(aioclient_mock)
        entity = (await self.async_entities(aioclient_mock))[0]

        assert entity.name == "My wLightBoxS (wLightBoxS#brightness_mono1)"
        assert entity.unique_id == "BleBox-wLightBoxS-1afe34e750b8-brightness_mono1"
        assert entity.brightness == 0xF5
        assert entity.is_on
        assert entity._feature.index == 0

    async def test_device_info(self, aioclient_mock):
        await self.allow_get_info(aioclient_mock, self.DEVICE_INFO)
        entity = (await self.async_entities(aioclient_mock))[0]
        assert entity.device_info["name"] == "My wLightBoxS"
        assert entity.device_info["mac"] == "1afe34e750b8"
        assert entity.device_info["manufacturer"] == "BleBox"
        assert entity.device_info["model"] == "wLightBoxS"
        assert entity.device_info["sw_version"] == "0.924"

    async def test_device_info2(self, aioclient_mock):
        await self.allow_get_info(aioclient_mock, self.DEVICE_INFO2)
        entity = (await self.async_entities(aioclient_mock))[0]
        assert entity.device_info["name"] == "My wLightBoxS"
        assert entity.device_info["mac"] == "ce50e32d2707"
        assert entity.device_info["manufacturer"] == "BleBox"
        assert entity.device_info["model"] == "wLightBoxS"
        assert entity.device_info["sw_version"] == "0.1022"

    async def test_update(self, aioclient_mock):
        """Test light updating."""

        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT)
        assert entity.brightness == 0xAB
        assert entity.is_on is True

    async def allow_set_brightness(self, code, aioclient_mock, value, response):
        """Set up mock for HTTP POST simulating color change."""
        raw = "{:02x}".format(value)
        await self.allow_post(
            code,
            aioclient_mock,
            "/api/rgbw/set",
            json.dumps(
                {"rgbw": {"desiredColor": raw + "------"}}
            ),  # simulating mask for color mod 3
            response,
        )

    async def test_on(self, aioclient_mock):
        """Test light on."""
        entity = await self.updated(aioclient_mock, self.STATE_OFF)
        assert entity.is_on is False

        async def turn_on():
            await entity.async_turn_on()

        await self.allow_set_brightness(
            turn_on, aioclient_mock, 0xFF, self.STATE_FULL_ON
        )

        assert entity.is_on is True
        assert entity.brightness == 0xFF

    async def test_on_with_bad_value_type(self, aioclient_mock):
        """Test light on with off value."""
        entity = await self.updated(aioclient_mock, self.STATE_OFF)
        assert entity.is_on is False

        with pytest.raises(
            ValueError,
            match=r"adjust_brightness called with bad parameter \(00 is <class 'str'> instead of int\)",
        ):
            await entity.async_turn_on(brightness="00")

    async def test_on_with_bad_value_exceeding_max(self, aioclient_mock):
        """Test light on with off value."""
        entity = await self.updated(aioclient_mock, self.STATE_OFF)
        assert entity.is_on is False

        with pytest.raises(
            ValueError,
            match=r"adjust_brightness called with bad parameter \(1234 is greater than 255\)",
        ):
            await entity.async_turn_on(brightness=1234)

    async def test_off(self, aioclient_mock):
        """Test light off."""
        entity = await self.updated(aioclient_mock, self.STATE_ON)
        assert entity.is_on is True

        async def turn_off():
            await entity.async_turn_off()

        await self.allow_set_brightness(turn_off, aioclient_mock, 0, self.STATE_OFF)

        assert entity.is_on is False
        assert entity.brightness == 0

    async def test_unspecified_version(self, aioclient_mock):
        """
        As default_api_level implemented for this class of devices, test method shall be omitted.
        """
        pass


class TestWLightBox(DefaultBoxTest):
    """Tests for BleBox wLightBox."""

    DEVCLASS = "lights"
    ENTITY_CLASS = BleBoxLightEntity

    DEVICE_EXTENDED_INFO_PATH = "/api/rgbw/extended/state"
    # TODO: rename everywhere (STATE_PATH)
    DEV_INFO_PATH = "api/rgbw/state"

    DEVICE_INFO = json.loads(
        """
    {
        "device": {
            "deviceName": "My light 1",
            "type": "wLightBox",
            "fv": "0.993",
            "hv": "4.3",
            "id": "1afe34e750b8",
            "apiLevel": 20190808
        }
    }
    """
    )
    DEVICE_EXTENDED_INFO = {
        "rgbw": {
            "colorMode": 4,
            "effectID": 0,
            "desiredColor": "fa00203A",
            "currentColor": "ff00302F",
            "lastOnColor": "f1e2d3e4",
            "durationsMs": {"colorFade": 1000, "effectFade": 1500, "effectStep": 2000},
            "favColors": {"0": "ff", "1": "00", "2": "c0", "3": "40", "4": "00"},
            "effectsNames": {"0": "NONE", "1": "FADE", "2": "Stroboskop", "3": "BELL"},
        }
    }
    DEVICE_EXTENDED_INFO_COLORMODE_1 = {
        "rgbw": {
            "colorMode": 1,
            "effectID": 0,
            "desiredColor": "fa",
            "currentColor": "ff",
            "lastOnColor": "ff",
            "durationsMs": {"colorFade": 1000, "effectFade": 1500, "effectStep": 2000},
            "favColors": {"0": "ff", "1": "00", "2": "c0", "3": "40", "4": "00"},
            "effectsNames": {"0": "NONE", "1": "FADE", "2": "Stroboskop", "3": "BELL"},
        }
    }
    DEVICE_EXTENDED_INFO_COLORMODE_2 = {
        "rgbw": {
            "colorMode": 2,
            "effectID": 0,
            "desiredColor": "fa00203A",
            "currentColor": "ff00302F",
            "lastOnColor": "f1e2d3e4",
            "durationsMs": {"colorFade": 1000, "effectFade": 1500, "effectStep": 2000},
            "favColors": {"0": "ff", "1": "00", "2": "c0", "3": "40", "4": "00"},
            "effectsNames": {"0": "NONE", "1": "FADE", "2": "Stroboskop", "3": "BELL"},
        }
    }
    DEVICE_EXTENDED_INFO_COLORMODE_3 = {
        "rgbw": {
            "colorMode": 3,
            "effectID": 0,
            "desiredColor": "fa00203A",
            "currentColor": "ff00302F",
            "lastOnColor": "f1e2d3e4",
            "durationsMs": {"colorFade": 1000, "effectFade": 1500, "effectStep": 2000},
            "favColors": {"0": "ff", "1": "00", "2": "c0", "3": "40", "4": "00"},
            "effectsNames": {"0": "NONE", "1": "FADE", "2": "Stroboskop", "3": "BELL"},
        }
    }
    DEVICE_EXTENDED_INFO_COLORMODE_4 = {
        "rgbw": {
            "colorMode": 4,
            "effectID": 0,
            "desiredColor": "fa00203A",
            "currentColor": "ff00302F",
            "lastOnColor": "f1e2d3e4",
            "durationsMs": {"colorFade": 1000, "effectFade": 1500, "effectStep": 2000},
            "favColors": {"0": "ff", "1": "00", "2": "c0", "3": "40", "4": "00"},
            "effectsNames": {"0": "NONE", "1": "FADE", "2": "Stroboskop", "3": "BELL"},
        }
    }
    DEVICE_EXTENDED_INFO_COLORMODE_5 = {
        "rgbw": {
            "colorMode": 5,
            "effectID": 0,
            "desiredColor": "fa00203A",
            "currentColor": "ff00302F",
            "lastOnColor": "f1e2d3e4",
            "durationsMs": {"colorFade": 1000, "effectFade": 1500, "effectStep": 2000},
            "favColors": {"0": "ff", "1": "00", "2": "c0", "3": "40", "4": "00"},
            "effectsNames": {"0": "NONE", "1": "FADE", "2": "Stroboskop", "3": "BELL"},
        }
    }

    DEVICE_EXTENDED_INFO_COLORMODE_6 = {
        "rgbw": {
            "colorMode": 6,
            "effectID": 0,
            "desiredColor": "fa00203A",
            "currentColor": "ff00302F",
            "lastOnColor": "f1e2d3e4",
            "durationsMs": {"colorFade": 1000, "effectFade": 1500, "effectStep": 2000},
            "favColors": {"0": "ff", "1": "00", "2": "c0", "3": "40", "4": "00"},
            "effectsNames": {"0": "NONE", "1": "FADE", "2": "Stroboskop", "3": "BELL"},
        }
    }
    DEVICE_EXTENDED_INFO_COLORMODE_7 = {
        "rgbw": {
            "colorMode": 7,
            "effectID": 0,
            "desiredColor": "fcfffcff00",
            "currentColor": "fcfffcff00",
            "lastOnColor": "fcfffcff00",
            "durationsMs": {"colorFade": 1000, "effectFade": 1000, "effectStep": 1000},
            "favColors": {"0": "ff", "1": "00", "2": "c0", "3": "40", "4": "00"},
            "effectsNames": {"0": "NONE", "1": "FADE", "2": "Stroboskop", "3": "BELL"},
        }
    }

    def patch_version(apiLevel):
        """Generate a patch for a JSON state fixture."""
        return f"""
        {{ "device": {{ "apiLevel": {apiLevel} }} }}
        """

    DEVICE_INFO_FUTURE = jmerge(DEVICE_INFO, patch_version(future_date()))
    DEVICE_INFO_LATEST = jmerge(
        DEVICE_INFO, patch_version(get_latest_api_level("wLightBox"))
    )
    DEVICE_INFO_UNSUPPORTED = jmerge(DEVICE_INFO, patch_version(20140717))

    DEVICE_INFO_UNSPECIFIED_API = json.loads(
        """
    {
        "device": {
            "deviceName": "My light 1",
            "type": "wLightBox",
            "fv": "0.247",
            "hv": "0.2",
            "id": "1afe34e750b8"
        },
        "network": {
            "ip": "192.168.1.237",
            "ssid": "myWiFiNetwork",
            "station_status": 5,
            "apSSID": "wLightBox-ap",
            "apPasswd": ""
        },
        "rgbw": {
            "desiredColor": "abcdefd9",
            "currentColor": "abcdefd9",
            "fadeSpeed": 248,
            "effectSpeed": 2,
            "effectID": 3,
            "colorMode": 3
        }
    }
            """
    )

    STATE_DEFAULT = json.loads(
        """
    {
      "rgbw": {
        "colorMode": 4,
        "effectID": 0,
        "desiredColor": "fa00203A",
        "currentColor": "ff00302F",
        "lastOnColor": "f1e2d3e4",
        "durationsMs": {
          "colorFade": 1000,
          "effectFade": 1500,
          "effectStep": 2000
        }
      }
    }
    """
    )

    def patch_state(current, desired=None, last=None):
        """Generate a patch for a JSON state fixture."""

        if desired is None:
            desired = current

        return f"""
        {{
            "rgbw": {{
                "currentColor": "{current}",
                "desiredColor": "{desired}"
                { "" if last is None else f',"lastOnColor": "{last}"' }
            }}
        }}
        """

    STATE_OFF = jmerge(STATE_DEFAULT, patch_state("00000000"))
    STATE_OFF_NOLAST_WHITE = jmerge(
        STATE_DEFAULT, patch_state("0a0b0c0d", "00000000", "dacefb00")
    )
    STATE_ON = STATE_DEFAULT
    STATE_ON_AFTER_WHITE = jmerge(
        STATE_DEFAULT, patch_state("01020304", "f1e2d3c7", "f1e2d3c7")
    )
    STATE_ON_AFTER_RESET_WHITE = jmerge(
        STATE_DEFAULT, patch_state("01020304", "f1e2d300", "f1e2d300")
    )

    STATE_ON_ONLY_SOME_COLOR = jmerge(STATE_DEFAULT, patch_state("ffa1b200"))
    STATE_ON_LAST = jmerge(STATE_DEFAULT, patch_state("01020304", "f1e2d3e4"))
    STATE_AFTER_SOME_COLOR_SET = jmerge(STATE_DEFAULT, patch_state("ffa1b2e4"))

    async def test_init(self, aioclient_mock):
        """Test cover default state."""
        await self.allow_get_info(aioclient_mock)
        entity = (await self.async_entities(aioclient_mock))[0]

        assert entity.name == "My light 1 (wLightBox#color_RGBorW)"
        assert entity.unique_id == "BleBox-wLightBox-1afe34e750b8-color_RGBorW"
        # In current state of master branch white_value is not property of BleBoxLightEntity, fake test... dissapointing
        assert entity._feature.supports_white
        assert entity._feature.white_value == 0x3A

        assert entity._feature.supports_color
        assert entity.brightness == 0xFA
        assert entity.is_on is True

    async def test_device_info(self, aioclient_mock):
        await self.allow_get_info(aioclient_mock, self.DEVICE_INFO)
        entity = (await self.async_entities(aioclient_mock))[0]
        assert entity.device_info["name"] == "My light 1"
        assert entity.device_info["mac"] == "1afe34e750b8"
        assert entity.device_info["manufacturer"] == "BleBox"
        assert entity.device_info["model"] == "wLightBox"
        assert entity.device_info["sw_version"] == "0.993"

    async def test_update(self, aioclient_mock):
        """Test light updating."""

        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT)
        assert entity.brightness == 250
        # assert entity.hs_color == (352.32, 100.0)
        # assert entity.white_value == 0x3A
        assert entity.is_on is True  # state already available

    async def allow_set_color(self, code, aioclient_mock, value, response):
        """Set up mock for HTTP POST simulating color change."""
        await self.allow_post(
            code,
            aioclient_mock,
            "/api/rgbw/set",
            '{"rgbw":{"desiredColor": "' + str(value) + '"}}',
            response,
        )

    async def test_on_to_last_color(self, aioclient_mock):
        """Test light on."""
        entity = await self.updated(aioclient_mock, self.STATE_OFF)
        assert entity.is_on is False

        async def action():
            await entity.async_turn_on()

        await self.allow_set_color(
            action, aioclient_mock, "f1e2d3e4", self.STATE_ON_LAST
        )

        assert entity.is_on is True
        # assert entity.white_value == 0xE4
        assert entity.rgbw_color == tuple(
            [int(i, 16) for i in ["f1", "e2", "d3", "e4"]]
        )

    async def test_off(self, aioclient_mock):
        """Test light off."""
        entity = await self.updated(aioclient_mock, self.STATE_ON)
        assert entity.is_on is True

        async def action():
            await entity.async_turn_off()

        await self.allow_set_color(action, aioclient_mock, "00000000", self.STATE_OFF)

        assert entity.is_on is False
        # assert entity.white_value == 0x00

    async def test_colormode_5_brightness(self, aioclient_mock):
        self.DEVICE_EXTENDED_INFO = self.DEVICE_EXTENDED_INFO_COLORMODE_5
        await self.allow_get_info(aioclient_mock)
        self.STATE_DEFAULT["colorMode"] = 5
        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT)

        assert "_cct" in entity.name
        assert entity.brightness == 250

    async def test_colormode_6_brightness(self, aioclient_mock):
        self.DEVICE_EXTENDED_INFO = self.DEVICE_EXTENDED_INFO_COLORMODE_6
        await self.allow_get_info(aioclient_mock)
        self.STATE_DEFAULT["colorMode"] = 6
        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT, 1)

        assert "_cct2" in entity.name
        assert entity.brightness
        assert entity._feature.index == 1

    async def test_effect_list_return_list(self, aioclient_mock):
        self.DEVICE_EXTENDED_INFO = self.DEVICE_EXTENDED_INFO_COLORMODE_5
        await self.allow_get_info(aioclient_mock)
        self.STATE_DEFAULT["colorMode"] = 5
        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT)

        assert entity.effect_list

    async def test_color_temp_for_colomode_6(self, aioclient_mock):
        self.DEVICE_EXTENDED_INFO = self.DEVICE_EXTENDED_INFO_COLORMODE_6
        await self.allow_get_info(aioclient_mock)
        self.STATE_DEFAULT["colorMode"] = 6
        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT)

        assert "_cct1" in entity.name
        assert entity.color_temp
        assert entity._feature.index == 0

    async def test_mono_4ch_indices(self, aioclient_mock):
        """Test that 4-channel MONO mode produces entities with correct indices."""
        self.DEVICE_EXTENDED_INFO = self.DEVICE_EXTENDED_INFO_COLORMODE_3
        await self.allow_get_info(aioclient_mock)
        self.STATE_DEFAULT["rgbw"]["colorMode"] = 3
        entities = await self.async_entities(aioclient_mock)

        assert len(entities) == 4
        for i, entity in enumerate(entities):
            assert entity._feature.index == i

    async def test_color_temp_for_colomode_rgbww(self, aioclient_mock):
        self.DEVICE_EXTENDED_INFO = self.DEVICE_EXTENDED_INFO_COLORMODE_7
        await self.allow_get_info(aioclient_mock)
        self.STATE_DEFAULT["colorMode"] = 7
        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT)

        assert "_RGBCCT" in entity.name
        assert entity.color_temp
        assert entity.brightness

    async def test_normalise_element_colormode_rgb(self, aioclient_mock):
        # testing sensible on value which is used only while async_turn_on executed

        self.DEVICE_EXTENDED_INFO = self.DEVICE_EXTENDED_INFO_COLORMODE_2
        self.DEVICE_EXTENDED_INFO = jmerge(
            self.DEVICE_EXTENDED_INFO, self.patch_state("fafafa", "fafafa")
        )
        self.STATE_DEFAULT["rgbw"]["colorMode"] = 2

        self.STATE_DEFAULT = jmerge(
            self.STATE_DEFAULT, self.patch_state("fafafa", "fafafa")
        )

        await self.allow_get_info(aioclient_mock)
        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT)

        async def turn_on():
            await entity.async_turn_on(rgb_color=(255, 0, 140))

        self.STATE_ON = jmerge(self.STATE_ON, self.patch_state("fafafa", "fafafa"))
        await self.allow_set_color(turn_on, aioclient_mock, "fa0089", self.STATE_ON)

        assert max(entity.rgbw_color) == 250

    async def test_normalise_when_max_is_zero_rgb(self, aioclient_mock):
        # testing sensible on value which is used only while async_turn_on executed

        self.DEVICE_EXTENDED_INFO = self.DEVICE_EXTENDED_INFO_COLORMODE_2
        self.DEVICE_EXTENDED_INFO = jmerge(
            self.DEVICE_EXTENDED_INFO, self.patch_state("030303", "030303")
        )
        self.STATE_DEFAULT["rgbw"]["colorMode"] = 2

        self.STATE_DEFAULT = jmerge(
            self.STATE_DEFAULT, self.patch_state("030303", "030303")
        )

        await self.allow_get_info(aioclient_mock)
        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT)

        async def turn_on():
            await entity.async_turn_on(rgb_color=(255, 10, 255))

        self.STATE_ON = jmerge(self.STATE_ON, self.patch_state("000000", "000000"))
        await self.allow_set_color(turn_on, aioclient_mock, "030003", self.STATE_ON)

        assert max(entity.rgb_color) == 255

    async def test_sensible_on_value_for_color_mode_1(self, aioclient_mock):
        self.DEVICE_EXTENDED_INFO = self.DEVICE_EXTENDED_INFO_COLORMODE_1
        self.DEVICE_EXTENDED_INFO = jmerge(
            self.DEVICE_EXTENDED_INFO, self.patch_state("00000000", "00000000")
        )
        self.STATE_DEFAULT["rgbw"]["colorMode"] = 1

        self.STATE_DEFAULT = jmerge(
            self.STATE_DEFAULT, self.patch_state("00000000", "00000000")
        )

        await self.allow_get_info(aioclient_mock)

        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT)

        async def turn_on():
            await entity.async_turn_on()

        self.STATE_ON = jmerge(self.STATE_ON, self.patch_state("ffffffff", "ffffffff"))
        await self.allow_set_color(turn_on, aioclient_mock, "ffffffff", self.STATE_ON)

        assert entity.rgbw_color == (255, 255, 255, 255)

    async def test_sensible_on_value_for_color_mode_5(self, aioclient_mock):
        self.DEVICE_EXTENDED_INFO = self.DEVICE_EXTENDED_INFO_COLORMODE_5
        self.DEVICE_EXTENDED_INFO = jmerge(
            self.DEVICE_EXTENDED_INFO, self.patch_state("00000000", "00000000")
        )
        self.STATE_DEFAULT["rgbw"]["colorMode"] = 5

        self.STATE_DEFAULT = jmerge(
            self.STATE_DEFAULT, self.patch_state("00000000", "00000000")
        )

        await self.allow_get_info(aioclient_mock)
        self.STATE_DEFAULT["colorMode"] = 5
        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT)

        async def turn_on():
            await entity.async_turn_on()

        self.STATE_ON = jmerge(self.STATE_ON, self.patch_state("ffffffff", "ffffffff"))
        await self.allow_set_color(turn_on, aioclient_mock, "ffff----", self.STATE_ON)

        assert entity.color_temp == 128

    async def test_turn_on_color_temp_full_warm_for_color_mode_5(self, aioclient_mock):
        self.DEVICE_EXTENDED_INFO = self.DEVICE_EXTENDED_INFO_COLORMODE_5
        self.STATE_DEFAULT["rgbw"]["colorMode"] = 5

        await self.allow_get_info(aioclient_mock)
        self.STATE_DEFAULT["colorMode"] = 5
        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT)

        async def turn_on():
            await entity.async_turn_on(color_temp=1)

        self.STATE_ON = jmerge(self.STATE_ON, self.patch_state("02ffffff", "02ffffff"))

        await self.allow_set_color(turn_on, aioclient_mock, "02fa----", self.STATE_ON)

        assert entity.color_temp == 1

    async def test_turn_on_color_temp_full_cold_for_color_mode_5(self, aioclient_mock):
        self.DEVICE_EXTENDED_INFO = self.DEVICE_EXTENDED_INFO_COLORMODE_5
        self.STATE_DEFAULT["rgbw"]["colorMode"] = 5

        await self.allow_get_info(aioclient_mock)
        self.STATE_DEFAULT["colorMode"] = 5
        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT)

        async def turn_on():
            await entity.async_turn_on(color_temp=255)

        self.STATE_ON = jmerge(self.STATE_ON, self.patch_state("fa00ffff", "fa02ffff"))

        await self.allow_set_color(turn_on, aioclient_mock, "fa00----", self.STATE_ON)

        assert entity.color_temp == 255

    async def test_sensible_on_value_for_color_mode_6(self, aioclient_mock):
        self.DEVICE_EXTENDED_INFO = self.DEVICE_EXTENDED_INFO_COLORMODE_6
        self.STATE_DEFAULT["rgbw"]["colorMode"] = 6

        await self.allow_get_info(aioclient_mock)
        self.STATE_DEFAULT["colorMode"] = 6
        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT)

        async def turn_on():
            await entity.async_turn_on(color_temp=255)

        self.STATE_ON = jmerge(self.STATE_ON, self.patch_state("fa00ffff", "fa02ffff"))

        await self.allow_set_color(turn_on, aioclient_mock, "fa00----", self.STATE_ON)

        assert entity.color_temp == 255

    async def test_sensible_on_value_for_color_mode_7(self, aioclient_mock):
        self.DEVICE_EXTENDED_INFO = self.DEVICE_EXTENDED_INFO_COLORMODE_7
        self.STATE_DEFAULT["rgbw"]["colorMode"] = 7
        self.STATE_DEFAULT = jmerge(
            self.STATE_DEFAULT, self.patch_state("fcfffcff00", "fcfffcff00")
        )
        await self.allow_get_info(aioclient_mock)
        self.STATE_DEFAULT["colorMode"] = 7
        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT)

        async def turn_on():
            await entity.async_turn_on(rgbww_color=(0, 0, 0, 120, 214))

        self.STATE_ON = jmerge(
            self.STATE_ON, self.patch_state("000000d678", "000000d678")
        )

        await self.allow_set_color(turn_on, aioclient_mock, "000000d678", self.STATE_ON)

        assert entity.rgbww_color == (0, 0, 0, 120, 214)

    async def test_unspecified_version(self, aioclient_mock):
        """
        As default_api_level implemented for this class of devices, test method shall be omitted.
        """
        pass


def test_unit_light_evaluate_brightness_from_rgb():
    tested_ob = Light.evaluate_brightness_from_rgb(iterable=(140, 230))
    assert tested_ob == 230


def test_unit_light_evaluate_off_value_mono_2ch():
    """Test evaluate_off_value for MONO mode with 2 channels (e.g., desired_color="0000")."""
    from blebox_uniapi.box import Box
    from unittest.mock import MagicMock

    box = MagicMock(spec=Box)
    box.type = "wLightBox"

    def mask(x):
        return f"{x}------"

    light = Light(
        product=box,
        alias="test_mono1",
        methods={},
        extended_state={"rgbw": {"colorMode": 3}},
        mask=mask,
        desired_color="0000",
        color_mode=3,
        effect_list=None,
        current_effect=None,
    )

    assert light._off_value == "00"


def test_unit_light_evaluate_off_value_mono_4ch():
    """Test evaluate_off_value for MONO mode with 4 channels (e.g., desired_color="00000000")."""
    from blebox_uniapi.box import Box
    from unittest.mock import MagicMock

    box = MagicMock(spec=Box)
    box.type = "wLightBox"

    def mask(x):
        return f"------{x}"

    light = Light(
        product=box,
        alias="test_mono4",
        methods={},
        extended_state={"rgbw": {"colorMode": 3}},
        mask=mask,
        desired_color="00000000",
        color_mode=3,
        effect_list=None,
        current_effect=None,
    )

    assert light._off_value == "00"


def test_unit_light_evaluate_off_value_ct():
    """Test evaluate_off_value for CT (color temperature) mode."""
    from blebox_uniapi.box import Box
    from unittest.mock import MagicMock

    box = MagicMock(spec=Box)
    box.type = "wLightBox"

    def mask(x):
        return f"{x}----"

    light = Light(
        product=box,
        alias="test_cct",
        methods={},
        extended_state={"rgbw": {"colorMode": 5}},
        mask=mask,
        desired_color="ffff",
        color_mode=5,
        effect_list=None,
        current_effect=None,
    )

    assert light._off_value == "0000"


def test_unit_light_get_placeholder_for_color_mode():
    """Test _get_placeholder_for_color_mode returns correct placeholder for each color mode."""
    from blebox_uniapi.box import Box
    from blebox_uniapi.light import BleboxColorMode
    from unittest.mock import MagicMock

    box = MagicMock(spec=Box)
    box.type = "wLightBox"

    def mask(x):
        return x

    test_cases = [
        (BleboxColorMode.MONO, "xx"),
        (BleboxColorMode.CT, "xxxx"),
        (BleboxColorMode.CTx2, "xxxx"),
        (BleboxColorMode.RGB, "xxxxxx"),
        (BleboxColorMode.RGBW, "xxxxxxxx"),
        (BleboxColorMode.RGBorW, "xxxxxxxx"),
    ]

    for color_mode, expected_placeholder in test_cases:
        light = Light(
            product=box,
            alias=f"test_{color_mode}",
            methods={},
            extended_state={"rgbw": {"colorMode": color_mode}},
            mask=mask,
            desired_color="ff000000",
            color_mode=color_mode,
            effect_list=None,
            current_effect=None,
        )
        assert light._get_placeholder_for_color_mode() == expected_placeholder


def test_unit_light_set_last_on_value_with_last_color():
    """Test _set_last_on_value calls expect_rgbw when last_color exists and raw is off."""
    from blebox_uniapi.box import Box
    from unittest.mock import MagicMock

    box = MagicMock(spec=Box)
    box.type = "wLightBox"
    box.expect_rgbw = MagicMock(return_value="ff00ff00")
    box.last_data = {"rgbw": {"lastOnColor": "ff00ff00"}}

    def mask(x):
        return x

    light = Light(
        product=box,
        alias="test_last_color",
        methods={"last_color": "rgbw.lastOnColor"},
        extended_state={"rgbw": {"colorMode": 4, "lastOnColor": "ff00ff00"}},
        mask=mask,
        desired_color="00000000",
        color_mode=4,
        effect_list=None,
        current_effect=None,
    )

    light._set_last_on_value("test_alias", box, light._off_value)
    box.expect_rgbw.assert_called_once_with("test_alias", "ff00ff00")
    assert light._last_on_state == "ff00ff00"


def test_unit_light_set_last_on_value_without_last_color():
    """Test _set_last_on_value uses default_on_value when last_color is None."""
    from blebox_uniapi.box import Box
    from unittest.mock import MagicMock

    box = MagicMock(spec=Box)
    box.type = "wLightBox"
    box.expect_rgbw = MagicMock(return_value="ff00ff00")
    box.last_data = {"rgbw": {}}

    def mask(x):
        return x

    light = Light(
        product=box,
        alias="test_no_last_color",
        methods={"last_color": "rgbw.lastOnColor"},
        extended_state={"rgbw": {"colorMode": 4}},
        mask=mask,
        desired_color="00000000",
        color_mode=4,
        effect_list=None,
        current_effect=None,
    )

    light._set_last_on_value("test_alias", box, light._off_value)
    box.expect_rgbw.assert_not_called()
    assert light._last_on_state == light._default_on_value


def test_unit_light_swap_ww_cw():
    """Test swap_ww_cw preserves first 3 channels and swaps last 2."""
    value = [100, 150, 200, 120, 80]
    result = Light.swap_ww_cw(value)
    assert result == [100, 150, 200, 80, 120]


def test_unit_light_swap_ww_cw_non_5_channel():
    """Test swap_ww_cw returns unchanged for non-5-channel values."""
    value = [100, 150, 200]
    result = Light.swap_ww_cw(value)
    assert result == [100, 150, 200]


def test_unit_light_scale_rgbww_brightness_preserves_ratio():
    """Test scale_rgbww_brightness preserves WW:CW ratio within rounding tolerance."""
    value = [100, 100, 100, 120, 80]
    result = Light.scale_rgbww_brightness(value, 100)
    expected_ratio = 120 / 80
    actual_ratio = result[3] / result[4]
    assert abs(actual_ratio - expected_ratio) < 0.01


def test_unit_light_scale_rgbww_brightness_zero():
    """Test scale_rgbww_brightness with zero max."""
    value = [0, 0, 0, 0, 0]
    result = Light.scale_rgbww_brightness(value, 200)
    assert result == [0, 0, 0, 0, 200]



from .feature import Feature
from typing import TYPE_CHECKING, Optional

from enum import Enum, auto

if TYPE_CHECKING:
    from .box import Box

TV_LIFT_CONTROL_TYPES_API = {
    0: {"1": "open_or_stop", "2": "close_or_stop"},
    1: {"1": "up_or_stop", "2": "down_or_stop"},
    2: {"1": "up_or_stop", "2": "down_or_stop"},
    3: {"1": "up_or_stop", "2": "down_or_stop"},
    4: {"1": "open_or_stop", "2": "close_or_stop", "3": "to_fav"},
}


class ControlType(Enum):
    UP = auto()
    DOWN = auto()
    FAVORITE = auto()
    OPEN = auto()
    CLOSE = auto()


class Button(Feature):
    def __init__(
        self, product: "Box", alias: str, methods: dict, query_string: str
    ) -> None:
        super().__init__(product, alias, methods)
        self._device_class = "UPDATE"
        self._query_string: str = query_string

    @classmethod
    def many_from_config(cls, product, box_type_config, extended_state):
        object_list = list()
        if len(box_type_config) > 0:
            alias = box_type_config[0]
            if isinstance(extended_state, dict) and extended_state is not None:
                lift_mode = extended_state.get("tvLift", {}).get("controlType", None)
                for row in TV_LIFT_CONTROL_TYPES_API[lift_mode].items():
                    indicator, endpoint = row
                    object_list.append(
                        cls(product, alias + "_" + endpoint, {}, endpoint)
                    )

                return object_list
        else:
            return []

    async def set(self):
        await self.async_api_command("set", self.query_string)

    def after_update(self) -> None:
        pass

    @property
    def control_type(self) -> Optional[ControlType]:
        """Return icon for endpoint."""
        if "up" in self.query_string:
            return ControlType.UP
        elif "down" in self.query_string:
            return ControlType.DOWN
        elif "fav" in self.query_string:
            return ControlType.FAVORITE
        elif "open" in self.query_string:
            return ControlType.OPEN
        elif "close" in self.query_string:
            return ControlType.CLOSE
        else:
            return None

    @property
    def query_string(self) -> str:
        return self._query_string



"""Blebox sensors tests."""

import json
import pytest

from blebox_uniapi.box_types import get_latest_api_level
from .conftest import CommonEntity, DefaultBoxTest, future_date, jmerge

TEMP_CELSIUS = "celsius"
DEVICE_CLASS_TEMPERATURE = "temperature class"


class BleBoxSensorEntity(CommonEntity):
    """Home Assistant representation style of a BleBox sensor feature."""

    @property
    def state(self):
        """Return the temperature."""
        return self._feature.current

    @property
    def unit_of_measurement(self):
        """Return the temperature unit."""
        return {"celsius": TEMP_CELSIUS}[self._feature.unit]

    @property
    def device_class(self):
        """Return the device class."""
        types = {"temperature": DEVICE_CLASS_TEMPERATURE}
        return types[self._feature.device_class]

    @property
    def native_value(self):
        """Return the state."""
        return self._feature.native_value


class TestTempSensor(DefaultBoxTest):
    """Tests for sensors representing BleBox tempSensor."""

    DEVCLASS = "sensors"
    ENTITY_CLASS = BleBoxSensorEntity

    DEV_INFO_PATH = "api/tempsensor/state"

    DEVICE_INFO = json.loads(
        """
    {
        "device": {
            "deviceName": "My tempSensor",
            "type": "tempSensor",
            "fv": "0.176",
            "hv": "0.6",
            "apiLevel": "20180604",
            "id": "1afe34db9437",
            "ip": "172.100.123.4"
        }
    }
    """
    )

    def patch_version(apiLevel):
        """Generate a patch for a JSON state fixture."""
        return f"""
        {{ "device": {{ "apiLevel": {apiLevel} }} }}
        """

    DEVICE_INFO_FUTURE = jmerge(DEVICE_INFO, patch_version(future_date()))
    DEVICE_INFO_LATEST = jmerge(
        DEVICE_INFO, patch_version(get_latest_api_level("tempSensor"))
    )
    DEVICE_INFO_UNSUPPORTED = jmerge(DEVICE_INFO, patch_version(20180603))

    DEVICE_INFO_UNSPECIFIED_API = json.loads(
        """
    {
        "device": {
            "deviceName": "My tempSensor",
            "type": "tempSensor",
            "fv": "0.176",
            "hv": "0.6",
            "id": "1afe34db9437",
            "ip": "172.100.123.4"
        }
    }
    """
    )

    STATE_DEFAULT = json.loads(
        """
    {
      "tempSensor": {
        "sensors": [
          {
            "type": "temperature",
            "id": 0,
            "value": 2518,
            "trend": 3,
            "state": 2,
            "elapsedTimeS": 0
          }
        ]
      }
    }
    """
    )

    DEVICE_EXTENDED_INFO_RAIN_AIR_TEMP = {}

    DEVICE_MULTISENSOR = json.loads(
        """
        {
            "multiSensor":
            {
                "sensors":
                [
                    {
                        "type": "temperature",
                        "id": 0,
                        "name": "Temperature outside",
                        "iconSet": 2,
                        "value": 3606,
                        "trend": 2,
                        "state": 2,
                        "elapsedTimeS": -1
                    },
                    {
                        "type": "temperature",
                        "id": 1,
                        "name": "Temperature inside fence",
                        "iconSet": 3,
                        "value": 4712,
                        "trend": 3,
                        "state": 2,
                        "elapsedTimeS": -1
                    },
                    {
                        "type": "temperature",
                        "id": 2,
                        "name": "Underground -10 cm",
                        "iconSet": 2,
                        "value": 2050,
                        "trend": 3,
                        "state": 2,
                        "elapsedTimeS": -1
                    }
                ]
            }
        }"""
    )

    DEVICE_MULTISENSOR_UPDATE = json.loads(
        """
        {
            "multiSensor":
            {
                "sensors":
                [
                    {
                        "type": "temperature",
                        "id": 0,
                        "name": "Temperature outside",
                        "iconSet": 2,
                        "value": 123,
                        "trend": 2,
                        "state": 2,
                        "elapsedTimeS": -1
                    },
                    {
                        "type": "temperature",
                        "id": 1,
                        "name": "Temperature inside fence",
                        "iconSet": 3,
                        "value": 213,
                        "trend": 3,
                        "state": 2,
                        "elapsedTimeS": -1
                    },
                    {
                        "type": "temperature",
                        "id": 2,
                        "name": "Underground -10 cm",
                        "iconSet": 2,
                        "value": 312,
                        "trend": 3,
                        "state": 2,
                        "elapsedTimeS": -1
                    }
                ]
            }
        }"""
    )
    DEVICE_EXTENDED_INFO = DEVICE_MULTISENSOR
    DEVICE_EXTENDED_INFO_PATH = "/state/extended"

    DEVICE_INFO_MULTISENSOR = json.loads(
        """
        {
          "device":
          {
            "deviceName": "Backyard - tempSensor PRO",
            "type": "multiSensor",
            "product": "tempSensorPro",
            "hv": "tSP-1.0",
            "fv": "0.1044",
            "universe": 0,
            "apiLevel": "20210413",
            "categories":
            [
              4,
              7
            ],
            "id": "42f5200ca102",
            "ip": "172.0.0.1",
            "availableFv": null
          }
        }
        """
    )

    async def test_init(self, aioclient_mock):
        """Test sensor default state."""

        await self.allow_get_info(aioclient_mock)
        entity = (await self.async_entities(aioclient_mock))[0]

        assert entity.device_class == DEVICE_CLASS_TEMPERATURE
        assert entity.unique_id == "BleBox-tempSensor-1afe34db9437-0.temperature"
        assert entity.unit_of_measurement == TEMP_CELSIUS
        assert entity.state is None
        # assert entity.outdated is False

    async def test_device_info(self, aioclient_mock):
        await self.allow_get_info(aioclient_mock, self.DEVICE_INFO)
        entity = (await self.async_entities(aioclient_mock))[0]
        assert entity.device_info["name"] == "My tempSensor"
        assert entity.device_info["mac"] == "1afe34db9437"
        assert entity.device_info["manufacturer"] == "BleBox"
        assert entity.device_info["model"] == "tempSensor"
        assert entity.device_info["sw_version"] == "0.176"

    async def test_update(self, aioclient_mock):
        """Test sensor update."""

        await self.allow_get_info(aioclient_mock)
        entity = (await self.async_entities(aioclient_mock))[0]

        self.allow_get_state(aioclient_mock, self.STATE_DEFAULT)
        await entity.async_update()

        # TODO: include product name?
        assert entity.name == "My tempSensor (tempSensor#0.temperature)"
        assert entity.state == 25.2
        assert entity._feature.index is None
        assert entity._feature.name is None

    async def test_sensor_factory(self, aioclient_mock):
        """Test sensor factory method class."""

        self.DEVICE_EXTENDED_INFO = self.DEVICE_MULTISENSOR
        self.STATE_DEFAULT = self.DEVICE_MULTISENSOR
        self.DEVICE_INFO = self.DEVICE_INFO_MULTISENSOR
        await self.allow_get_info(aioclient_mock)

        entities = await self.async_entities(aioclient_mock)

        assert len(entities) == 3
        assert entities[0]._feature.index == 0
        assert entities[0]._feature.name == "Temperature outside"
        assert entities[1]._feature.index == 1
        assert entities[1]._feature.name == "Temperature inside fence"
        assert entities[2]._feature.index == 2
        assert entities[2]._feature.name == "Underground -10 cm"

    async def test_multisensor_update(self, aioclient_mock):
        self.DEV_INFO_PATH = "state"
        self.DEVICE_EXTENDED_INFO = self.DEVICE_MULTISENSOR
        self.STATE_DEFAULT = self.DEVICE_MULTISENSOR
        self.DEVICE_INFO = self.DEVICE_INFO_MULTISENSOR
        # await self.allow_get_state(aioclient_mock)

        entity = await self.updated(aioclient_mock, self.DEVICE_MULTISENSOR_UPDATE)

        assert entity.native_value == 1.2


class TestMultiSensorEnergy(DefaultBoxTest):
    """Tests for multiSensor energy/current sensors (apiLevel 20230606)."""

    DEVCLASS = "sensors"
    ENTITY_CLASS = BleBoxSensorEntity

    DEV_INFO_PATH = "state"

    DEVICE_INFO = json.loads(
        """
        {
          "device": {
            "deviceName": "My multiSensor",
            "type": "multiSensor",
            "product": "energySensor",
            "hv": "eS-1.0",
            "fv": "0.1",
            "universe": 0,
            "apiLevel": "20230606",
            "id": "aa00112233bb",
            "ip": "172.0.0.1",
            "availableFv": null
          }
        }
        """
    )

    def patch_version(apiLevel):
        return f'{{ "device": {{ "apiLevel": {apiLevel} }} }}'

    DEVICE_INFO_FUTURE = jmerge(DEVICE_INFO, patch_version(future_date()))
    DEVICE_INFO_LATEST = jmerge(
        DEVICE_INFO, patch_version(get_latest_api_level("multiSensor"))
    )
    DEVICE_INFO_UNSUPPORTED = jmerge(DEVICE_INFO, patch_version(20180603))
    DEVICE_INFO_UNSPECIFIED_API = None

    STATE_DEFAULT = json.loads(
        """
        {
            "multiSensor": {
                "sensors": [
                    {"type": "current",             "id": 0, "value": 1500},
                    {"type": "forwardActiveEnergy",  "id": 1, "value": 5000},
                    {"type": "reverseActiveEnergy",  "id": 2, "value": 3000}
                ]
            }
        }
        """
    )

    DEVICE_EXTENDED_INFO = STATE_DEFAULT
    DEVICE_EXTENDED_INFO_PATH = "/state/extended"

    @pytest.mark.parametrize(
        "index,unit,value",
        [
            (0, "mA", 1500),
            (1, "kWh", 5.0),
            (2, "kWh", 3.0),
        ],
    )
    async def test_sensor_value(self, aioclient_mock, index, unit, value):
        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT, index=index)
        assert entity._feature.unit == unit
        assert entity.native_value == value


class TestAirSensor(DefaultBoxTest):
    """Tests for sensors representing BleBox airSensor."""

    DEV_INFO_PATH = "api/air/state"

    DEVCLASS = "sensors"
    ENTITY_CLASS = BleBoxSensorEntity

    DEVICE_INFO = json.loads(
        """
    {
        "deviceName": "My air 1",
        "type": "airSensor",
        "fv": "0.973",
        "hv": "0.6",
        "apiLevel": "20180403",
        "id": "1afe34db9437",
        "ip": "192.168.1.11"
    }
    """
    )

    def patch_version(apiLevel):
        """Generate a patch for a JSON state fixture."""
        return f"""
        {{ "apiLevel": {apiLevel} }}
        """

    DEVICE_INFO_FUTURE = jmerge(DEVICE_INFO, patch_version(future_date()))
    DEVICE_INFO_LATEST = jmerge(
        DEVICE_INFO, patch_version(get_latest_api_level("airSensor"))
    )
    DEVICE_INFO_UNSUPPORTED = jmerge(DEVICE_INFO, patch_version(20180402))

    DEVICE_INFO_UNSPECIFIED_API = json.loads(
        """
    {
        "deviceName": "My air 1",
        "type": "airSensor",
        "fv": "0.973",
        "hv": "0.6",
        "id": "1afe34db9437",
        "ip": "192.168.1.11"
    }
    """
    )

    STATE_DEFAULT = json.loads(
        """
        {
            "air": {
                "sensors": [
                    {
                        "type": "pm1",
                        "value": 49,
                        "trend": 3,
                        "state": 0,
                        "qualityLevel": 0,
                        "elaspedTimeS": -1
                    },
                    {
                        "type": "pm2.5",
                        "value": 222,
                        "trend": 1,
                        "state": 0,
                        "qualityLevel": 4,
                        "elaspedTimeS": -1
                    },
                    {
                        "type": "pm10",
                        "value": 333,
                        "trend": 0,
                        "state": 0,
                        "qualityLevel": 6,
                        "elaspedTimeS": -1
                    }
                ]
            }
        }
        """
    )

    # DEVICE_EXTENDED_INFO
    # DEVICE_EXTENDED_INFO_PATH

    async def test_init(self, aioclient_mock):
        """Test air quality sensor default state."""

        await self.allow_get_info(aioclient_mock)
        entity = (await self.async_entities(aioclient_mock))[0]

        assert entity.name == "My air 1 (airSensor#pm1)"
        assert entity.unique_id == "BleBox-airSensor-1afe34db9437-pm1"
        assert entity.native_value is None

    async def test_device_info(self, aioclient_mock):
        await self.allow_get_info(aioclient_mock, self.DEVICE_INFO)
        entity = (await self.async_entities(aioclient_mock))[0]
        assert entity.device_info["name"] == "My air 1"
        assert entity.device_info["mac"] == "1afe34db9437"
        assert entity.device_info["manufacturer"] == "BleBox"
        assert entity.device_info["model"] == "airSensor"
        assert entity.device_info["sw_version"] == "0.973"

    @pytest.mark.parametrize("io_param", [(0, 49), (1, 222), (2, 333)])
    async def test_update(self, aioclient_mock, io_param):
        """Test air quality sensor state after update."""

        await self.allow_get_info(aioclient_mock)
        entity = (await self.async_entities(aioclient_mock))[io_param[0]]

        self.allow_get_state(aioclient_mock, self.STATE_DEFAULT)
        await entity.async_update()

        assert entity.native_value == io_param[1]  # parametrised

    async def test_list_quantity(self, aioclient_mock):
        """Test air sensor init from config."""
        await self.allow_get_info(aioclient_mock)

        entity = await self.async_entities(aioclient_mock)

        assert len(entity) == 3



from enum import IntEnum
from .feature import Feature
from typing import TYPE_CHECKING, Optional, Dict, Any, Union, Sequence

if TYPE_CHECKING:
    from .box import Box

# V3 onwards used while device is operating on 5 channels, refactor if new versions occurs
ctx2_v3 = {"cct1": lambda x: f"{x}------", "cct2": lambda x: f"----{x}--"}
ctx2 = {"cct1": lambda x: f"{x}----", "cct2": lambda x: f"----{x}"}

mono = {
    "mono1": lambda x: f"{x}------",
    "mono2": lambda x: f"--{x}----",
    "mono3": lambda x: f"----{x}--",
    "mono4": lambda x: f"------{x}",
}


class BleboxColorMode(IntEnum):
    RGBW = 1  # RGB color-space with color brightness, white brightness
    RGB = 2  # RGB color-space with color brightness
    MONO = 3
    RGBorW = 4  # RGBW entity, where white color is prioritised
    CT = 5  # color-temperature, brightness, effect
    CTx2 = 6  # color-temperature, brightness, effect, two instances
    RGBWW = 7  # RGB with two color-temperature sliders(warm, cold)

    @classmethod
    def invert(cls):
        return {item.value: item.name for item in cls}


BLEBOX_COLOR_MODES = BleboxColorMode.invert()


class Light(Feature):
    # TODO: better defaults?
    CURRENT_CONF = dict()
    CONFIG = {
        "wLightBox": {
            "default": "FFFFFFFF",
            "off": "00000000",
            "white?": True,
            "color?": True,
            "to_value": lambda int_value: f"{int_value:02x}",
            "validator": lambda product, alias, raw: product.expect_rgbw(alias, raw),
        },
        "wLightBoxS": {
            "default": "FF",
            "off": "00",
            "white?": False,
            "color?": False,
            "to_value": lambda int_value: f"{int_value:02x}",
            "validator": lambda product, alias, raw: product.expect_hex_str(
                alias, raw, 255, 0
            ),
        },
        "dimmerBox": {
            "default": 0xFF,
            "off": 0x0,
            "white?": False,
            "color?": False,
            "to_value": lambda int_value: int_value,
            "validator": lambda product, alias, raw: product.expect_int(
                alias, raw, 255, 0
            ),
        },
    }

    COLOR_MODE_CONFIG = {
        "CT": {
            "default": "FFFFFFFF",
            "off": "0000",
            "white?": False,
            "color?": False,
            "to_value": lambda int_value: f"{int_value:02x}",
            "validator": lambda product, alias, raw: product.expect_hex_str(
                alias, raw, 255, 0
            ),
        },
        "CTx2": {
            "default": "FFFFFFFF",
            "off": "0000",
            "white?": False,
            "color?": False,
            "to_value": lambda int_value: f"{int_value:02x}",
            "validator": lambda product, alias, raw: product.expect_hex_str(
                alias, raw, 255, 0
            ),
        },
        "RGBWW": {
            "default": "FFFFFFFFFF",
            "off": "0000000000",
            "white?": True,
            "color?": True,
            "to_value": lambda int_value: f"{int_value:02x}",
            "validator": lambda product, alias, raw: product.expect_hex_str(
                alias, raw, 255, 0
            ),
        },
    }

    def __init__(
        self,
        product: "Box",
        alias: str,
        methods: dict,
        extended_state: Optional[Dict],
        mask: Any,
        desired_color,
        color_mode,
        effect_list,
        current_effect,
        index: Optional[int] = None,
    ) -> None:
        super().__init__(product, alias, methods)
        self._index = index
        config = self.CONFIG[product.type]
        self.mask = mask
        self.desired_color = desired_color
        self._color_mode = color_mode
        self._effect_list = effect_list
        if self._effect_list is not None:
            self._effect = current_effect

        if extended_state not in [None, {}]:
            self.extended_state = extended_state
            self.device_colorMode = color_mode
            if self.device_colorMode in [6, 7]:
                config = self.COLOR_MODE_CONFIG[
                    BLEBOX_COLOR_MODES[self.device_colorMode]
                ]
        else:
            if product.type == "dimmerBox":
                self.device_colorMode = BleboxColorMode.MONO
        self.CURRENT_CONF = config

        self._off_value = self.evaluate_off_value(config, desired_color)
        self._last_on_state = self._default_on_value = config["default"]

    @classmethod
    def many_from_config(
        cls, product, box_type_config, extended_state
    ) -> list["Light"]:
        if isinstance(extended_state, dict) and extended_state is not None:
            desired_color = extended_state.get("rgbw", {}).get("desiredColor")
            color_mode = extended_state.get("rgbw", {}).get("colorMode")
            current_effect = extended_state.get("rgbw", {}).get("effectID")
            effect_list = extended_state.get("rgbw", {}).get("effectsNames")
        else:
            desired_color = None
            color_mode = None
            current_effect = None
            effect_list = None
        alias, methods = box_type_config[0]
        const_kwargs = dict(
            methods=methods,
            extended_state=extended_state,
            desired_color=desired_color,
            color_mode=color_mode,
            current_effect=current_effect,
            effect_list=effect_list,
        )

        if extended_state is not None and color_mode is not None:
            if BleboxColorMode(color_mode).name == "RGBW":
                if len(desired_color) == 10:

                    def generate_mask(x):
                        return f"{x}--"

                    mask = generate_mask
                else:
                    mask = None
                return [cls(product, alias=alias + "_RGBW", mask=mask, **const_kwargs)]

            if BleboxColorMode(color_mode).name == "RGB":
                return [cls(product, alias=alias + "_RGB", mask=None, **const_kwargs)]

            if BleboxColorMode(color_mode).name == "MONO":
                if len(desired_color) % 2 == 0:
                    object_list = []
                    mono_item = list(mono.items())
                    for i in range(0, int(len(desired_color) / 2)):
                        indicator, mask = mono_item[i]
                        object_list.append(
                            cls(
                                product,
                                alias=alias + "_" + indicator,
                                mask=mask,
                                index=i,
                                **const_kwargs,
                            )
                        )
                    return object_list

            if BleboxColorMode(color_mode).name == "RGBorW":
                return [
                    cls(product, alias=alias + "_RGBorW", mask=None, **const_kwargs)
                ]

            if BleboxColorMode(color_mode).name == "CT":
                mask = ctx2["cct1"]
                if len(desired_color) > 8:
                    mask = ctx2_v3["cct1"]
                return [cls(product, alias=alias + "_cct", mask=mask, **const_kwargs)]

            if BleboxColorMode(color_mode).name == "CTx2":
                object_list = []
                ct = ctx2
                if len(desired_color) > 8:
                    ct = ctx2_v3
                for i, (indicator, mask) in enumerate(ct.items()):
                    object_list.append(
                        cls(
                            product,
                            alias=alias + "_" + indicator,
                            mask=mask,
                            index=i,
                            **const_kwargs,
                        )
                    )
                return object_list

            if BleboxColorMode(color_mode).name == "RGBWW":
                return [
                    cls(product, alias=alias + "_RGBCCT", mask=None, **const_kwargs)
                ]

        if len(box_type_config) > 0:
            del const_kwargs["methods"]
            if "Brightness" in box_type_config[0][1].get("desired"):
                const_kwargs["color_mode"] = BleboxColorMode.MONO
            return [
                cls(product, *args, mask=None, **const_kwargs)
                for args in box_type_config
            ]
        else:
            return []

    @property
    def brightness(self) -> Optional[int]:
        if self.color_mode in [6, 5]:
            _, bgt = self.color_temp_brightness_int_from_hex(self._desired)
            return bgt
        elif self.color_mode is not None and (
            rgb_list := self.rgb_hex_to_rgb_list(self.rgb_hex)
        ):
            return self.evaluate_brightness_from_rgb(rgb_list)
        else:
            return None

    @property
    def index(self) -> Optional[int]:
        return self._index

    @property
    def effect_list(self):
        if isinstance(self._effect_list, dict):
            return list(self._effect_list.values())
        else:
            return []

    @property
    def color_temp(self):
        ct, _ = self.color_temp_brightness_int_from_hex(self._desired)
        return ct

    @staticmethod
    def evaluate_brightness_from_rgb(iterable: Sequence[int]) -> int:
        "return brightness from 0 to 255 evaluated basing rgb"
        if max(iterable) > 255:
            raise ValueError(
                f"evaluate_brightness_from_rgb values out of range, max is {max(iterable)}."
            )
        elif min(iterable) < 0:
            raise ValueError(
                f"evaluate_brightness_from_rgb values out of range, min is {min(iterable)}."
            )
        return int(max(iterable))

    @staticmethod
    def swap_ww_cw(value: list) -> list:
        """Swap WW and CW channels in RGBWW mode for API compatibility."""
        if len(value) == 5:
            value = list(value)
            value[3], value[4] = value[4], value[3]
        return value

    @staticmethod
    def scale_rgbww_brightness(value: list, brightness: int) -> list:
        """Scale RGBWW value preserving WW:CW ratio."""
        current_max = max(value)
        if current_max > 0:
            return [round(x * brightness / current_max) for x in value]
        else:
            return [0, 0, 0, 0, brightness]

    def apply_brightness(self, value: int, brightness: int) -> Any:
        """Return list of values with applied brightness."""
        if not isinstance(brightness, int):
            raise ValueError(
                f"adjust_brightness called with bad parameter ({brightness} is {type(brightness)} instead of int)"
            )

        if brightness > 255:
            raise ValueError(
                f"adjust_brightness called with bad parameter ({brightness} is greater than 255)"
            )

        if self.product.type == "dimmerBox" or self.color_mode == BleboxColorMode.MONO:
            return [brightness]
        if brightness == 0:
            return [value]

        if self.color_mode in (
            BleboxColorMode.RGB,
            BleboxColorMode.RGBW,
            BleboxColorMode.RGBorW,
        ):
            value = self.normalise_elements_of_rgb(list(value))
            res = list(map(lambda x: round(x * (brightness / 255)), value))
        elif self.color_mode == BleboxColorMode.RGBWW:
            res = self.scale_rgbww_brightness(value, brightness)
        else:
            res = list(map(lambda x: round(x * (brightness / 255)), value))
        return res

    def evaluate_off_value(self, config: dict, raw_hex: str) -> str:
        """
        Return hex representing off state value without mask formatting for necessary channels if mask is applied.
        If no mask applied than return default from config

        :param config:
        :param raw_hex:
        :return: str
        """
        if self.mask:
            placeholder = self._get_placeholder_for_color_mode()
            mask_result = self.mask(placeholder)
            x_count = mask_result.count("x")
            return "0" * x_count
        elif raw_hex is not None:
            if len(raw_hex) < len(config["off"]):
                return config["off"][: len(raw_hex)]
        return config["off"]

    @property
    def supports_white(self) -> Any:
        return self.CURRENT_CONF["white?"]

    @property
    def white_value(self) -> Optional[int]:
        return self._white_value

    def apply_white(self, value: str, white: int) -> Union[int, str]:
        if white is None:
            return value

        if not self.supports_white:
            return value

        rgbhex = value[0:6]
        white_raw = f"{white:02x}"
        return f"{rgbhex}{white_raw}"

    @property
    def supports_color(self) -> Any:
        return self.CURRENT_CONF["color?"]

    @property
    def color_mode(self) -> int:
        return self._color_mode

    def apply_color(self, value: str, rgb_hex: str) -> Union[int, str]:
        if rgb_hex is None:
            return value

        if not self.supports_color:
            return value

        white_hex = value[6:8]
        return f"{rgb_hex}{white_hex}"

    def _get_placeholder_for_color_mode(self) -> str:
        if self._color_mode == BleboxColorMode.MONO:
            return "xx"
        elif self._color_mode in (BleboxColorMode.CT, BleboxColorMode.CTx2):
            return "xxxx"
        elif self._color_mode == BleboxColorMode.RGB:
            return "xxxxxx"
        elif self._color_mode in (BleboxColorMode.RGBW, BleboxColorMode.RGBorW):
            return "xxxxxxxx"
        else:
            return "x"

    def return_color_temp_with_brightness(
        self, value, brightness: Any
    ) -> Optional[str]:
        """Method returns value which will be send to"""
        if value < 128:
            warm = min(255, value * 2)
            cold = 255
        else:
            warm = 255
            cold = max(0, min(255, (255 - value) * 2))
        cold = cold * brightness / 255
        warm = warm * brightness / 255
        cold = f"{int(round(cold)):02x}"
        warm = f"{int(round(warm)):02x}"

        return self.rgb_hex_to_rgb_list(warm + cold)

    def value_for_selected_channels_from_given_val(self, value: str):
        placeholder = self._get_placeholder_for_color_mode()
        lambda_result = self.mask(placeholder)
        first_index = lambda_result.index("x")
        last_index = lambda_result.rindex("x")
        return value[first_index : last_index + 1]

    @staticmethod
    def color_temp_brightness_int_from_hex(val) -> (int, int):
        """Assuming that hex is 2channels, 4characters. Return values for front end"""

        cold = int(val[2:], 16)
        warm = int(val[0:2], 16)

        if cold > warm:
            if warm == 0:
                return 0, cold
            else:
                return round(int((128 * (warm * 255 / cold) / 255)), 2), cold
        if cold < warm:
            if cold == 0:
                return 255, warm
            else:
                return round(int(255 - 128 * (cold * (255 / warm) / 255)), 2), warm
        else:
            return 128, max(cold, warm)

    @staticmethod
    def normalise_elements_of_rgb(elements):
        max_val = max(elements)
        min_val = min(elements)
        if 0 > max_val or max_val > 255:
            raise ValueError(f"Max value in normalisation was outside range {max_val}.")
        elif min_val < 0:
            raise ValueError(f"Min value in normalisation was outside range {min_val}.")
        elif max_val == 0:
            return [255] * len(elements)
        return list(map(lambda x: round(x * 255 / max_val), elements))

    @property
    def is_on(self) -> Optional[bool]:
        return self._is_on

    @property
    def effect(self) -> Optional[str]:
        if isinstance(self._effect_list, dict):
            return self._effect_list.get(str(self._effect))
        return self._effect

    def after_update(self) -> None:
        alias = self._alias
        product = self._product

        if product.last_data is None:
            self._desired_raw = None
            self._desired = None
            self._is_on = None
            self._effect = None
            if self.mask is None:
                self._white_value = None
            return

        self._effect = self.raw_value("currentEffect")
        raw = self._return_desired_value(alias, product)
        self._set_last_on_value(alias, product, raw)
        self._set_is_on()

    def _set_last_on_value(self, alias, product, raw):
        if raw == self._off_value:
            last_color = self.raw_value("last_color")
            if last_color is not None:
                raw = product.expect_rgbw(alias, last_color)
                if self.mask is not None:
                    raw = self.value_for_selected_channels_from_given_val(raw)
                    if raw == self._off_value:
                        raw = self.value_for_selected_channels_from_given_val(
                            "ffffffffff"
                        )
                else:
                    if raw == self._off_value:
                        raw = "f" * len(raw)
            else:
                raw = self._default_on_value

        if raw in (self._off_value, None):
            raise ValueError(raw)
        # TODO: store as custom value permanently (exposed by API consumer)
        self._last_on_state = raw

    def _set_is_on(self):
        self._is_on = (self._off_value != self._desired) or (
            self._effect != 0 and self._effect is not None
        )
        if isinstance(self._desired, str):
            if int(self._desired, 16) == 0:
                self._is_on = False

    def _return_desired_value(self, alias, product) -> str:
        """
        Return value representing desired device state, set desired fields
        :param alias:
        :param product:
        :return desired value including mask:
        """

        response_desired_val = self.raw_value("desired")
        if self.mask is not None:
            raw = self.value_for_selected_channels_from_given_val(response_desired_val)
            self._desired = self.CONFIG[self._product.type]["validator"](
                product, alias, raw
            )
            if self.color_mode in [1, 4]:
                self._white_value = int(raw[6:8], 16)
        else:
            raw = response_desired_val
            self._desired_raw = raw
            self._desired = self.CONFIG[self._product.type]["validator"](
                product, alias, raw
            )  # type: ignore
            if self.color_mode in [1, 4]:
                self._white_value = int(raw[6:8], 16)
        return raw

    @property
    def sensible_on_value(self) -> Any:
        """Return sensible on value in hass format."""
        if self.mask is not None:
            if int(self._last_on_state, 16) == 0:
                if self.color_mode in (BleboxColorMode.RGBW, BleboxColorMode.RGBorW):
                    return 255, 255, 255, 255
                if self.color_mode == BleboxColorMode.MONO:
                    return [255]
                if self.color_mode in (BleboxColorMode.CT, BleboxColorMode.CTx2):
                    return 255, 255
            else:
                if self.color_mode in (
                    BleboxColorMode.MONO,
                    BleboxColorMode.CT,
                    BleboxColorMode.CTx2,
                ):
                    return self.rgb_hex_to_rgb_list(self._last_on_state)
                return self.normalise_elements_of_rgb(
                    self.rgb_hex_to_rgb_list(self._last_on_state)
                )
        else:
            if isinstance(self._last_on_state, str):
                if int(self._last_on_state, 16) == 0:
                    return [255] * len(self.rgb_hex_to_rgb_list(self._last_on_state))
                else:
                    if self.color_mode == BleboxColorMode.RGB:
                        return self.rgb_hex_to_rgb_list(self._last_on_state[:6])
                    elif self.color_mode == BleboxColorMode.MONO:
                        return self._last_on_state
                    elif self.color_mode == BleboxColorMode.RGBWW:
                        result = self.rgb_hex_to_rgb_list(self._last_on_state)
                        return self.swap_ww_cw(result)
                    else:
                        return self.rgb_hex_to_rgb_list(self._last_on_state)
            else:
                return self._last_on_state

    @property
    def rgb_hex(self) -> Any:
        """Return hex str representing rgb."""
        if isinstance(self._desired, int):
            return f"{self._desired:02x}"
        else:
            return self._desired

    @property
    def rgbw_hex(self) -> Any:
        return self._desired

    @property
    def rgbww_hex(self) -> Any:
        if len(self._desired) < 10:
            return None
        else:
            hex_str = self._desired
            hex_str_warm_cold = hex_str[6:]
            output_str = hex_str[0:6] + "".join(
                [
                    hex_str_warm_cold[i - 2 : i]
                    for i in range(len(hex_str_warm_cold), 0, -2)
                ]
            )
            return output_str

    @staticmethod
    def rgb_hex_to_rgb_list(hex_str) -> list[int]:
        """Return an RGB color value list from a hex color string."""
        if hex_str is not None:
            return [int(hex_str[i : i + 2], 16) for i in range(0, len(hex_str), 2)]
        return []

    @staticmethod
    def rgb_list_to_rgb_hex_list(rgb_list) -> hex:
        return [f"{i:02x}" for i in rgb_list]

    async def async_on(self, value: Any) -> None:
        if isinstance(value, (list, tuple)):
            if self.color_mode == BleboxColorMode.RGBWW:
                value.insert(3, value.pop())
            value = "".join(self.rgb_list_to_rgb_hex_list(value))
        if self.product.type == "dimmerBox":
            if not isinstance(value, int):
                value = int(value, 16)
        if not isinstance(value, type(self._off_value)):
            raise ValueError(
                f"turn_on called with bad parameter ({value} is {type(value)}, compared to {self._off_value} which is "
                f"{type(self._off_value)})"
            )

        if value == self._off_value:
            raise ValueError(f"turn_on called with invalid value ({value})")

        if self.mask is not None:
            value = self.mask(value)

        await self.async_api_command("set", value)

    async def async_off(self) -> None:
        if self.raw_value("colorMode") in [5, 6]:
            await self.async_api_command("set", self.mask("0000"))
        elif self.raw_value("colorMode") == 3 and self.product.type != "dimmerBox":
            await self.async_api_command("set", self.mask("00"))
        else:
            await self.async_api_command("set", self._off_value)



class Error(RuntimeError):
    """Generic blebox_uniapi error."""

    pass


# Likely fixable/retriable network/busy errors
class ConnectionError(Error):
    pass


class TimeoutError(ConnectionError):
    pass


# Likely unfixable device errors (do not setup)
class ClientError(Error):
    pass


class HttpError(ClientError):
    pass


class UnauthorizedRequest(ClientError):
    pass


# API errors
class BoxError(Error):
    pass


class UnsupportedBoxResponse(BoxError):
    pass


class UnsupportedBoxVersion(BoxError):
    pass



class BadFieldExceedsMax(BoxError):
    def __init__(self, dev_name: str, field: str, value: int, max_value: int):
        self._dev_name = dev_name
        self._field = field
        self._value = value
        self._max_value = max_value

    def __str__(self) -> str:
        return f"{self._dev_name}.{self._field} is {self._value} which exceeds max ({self._max_value})"


class BadFieldLessThanMin(BoxError):
    def __init__(self, dev_name: str, field: str, value: int, min_value: int):
        self._dev_name = dev_name
        self._field = field
        self._value = value
        self._min_value = min_value

    def __str__(self) -> str:
        return f"{self._dev_name}.{self._field} is {self._value} which is less than minimum ({self._min_value})"


class BadFieldMissing(BoxError):
    def __init__(self, dev_name: str, field: str):
        self._dev_name = dev_name
        self._field = field

    def __str__(self) -> str:
        return f"{self._dev_name}.{self._field} is missing"


class BadFieldNotANumber(BoxError):
    def __init__(self, dev_name: str, field: str, value: int):
        self._dev_name = dev_name
        self._field = field
        self._value = value

    def __str__(self) -> str:
        return (
            f"{self._dev_name}.{self._field} is '{self._value}' which is not a number"
        )


class BadFieldNotAString(BoxError):
    def __init__(self, dev_name: str, field: str, value: int):
        self._dev_name = dev_name
        self._field = field
        self._value = value

    def __str__(self) -> str:
        return f"{self._dev_name}.{self._field} is {self._value} which is not a string"


class BadFieldNotRGBW(BoxError):
    def __init__(self, dev_name: str, field: str, value: int):
        self._dev_name = dev_name
        self._field = field
        self._value = value

    def __str__(self) -> str:
        return f"{self._dev_name}.{self._field} is {self._value} which is not a rgbw string"



# misc errors


class MisconfiguredDevice(BoxError):
    pass


# development bugs that shouldn't normally be possible


class DeviceStateNotAvailable(BoxError):
    def __str__(self) -> str:
        return "device state not available yet"  # pragma: no cover



"""BleBox climate entities tests."""

import json

from blebox_uniapi.box_types import get_latest_api_level

from .conftest import CommonEntity, DefaultBoxTest, future_date, jmerge

# TODO: remove
SUPPORT_TARGET_TEMPERATURE = 1
HVAC_MODE_OFF = "hvac mode off"
HVAC_MODE_HEAT = "hvac mode heat"
CURRENT_HVAC_OFF = "current hvac mode off"
CURRENT_HVAC_HEAT = "current hvac mode heat"
CURRENT_HVAC_IDLE = "current hvac mode idle"
ATTR_TEMPERATURE = "temperature"
TEMP_CELSIUS = "celsius"


class ClimateDevice:
    def __init__(self):
        self._state = None

    @property
    def state(self):
        if self._feature.is_on is None:
            return None

        if not self._feature.is_on:
            return HVAC_MODE_OFF
        return HVAC_MODE_HEAT

    @property
    def device_class(self):
        return None


class BleBoxClimateEntity(CommonEntity, ClimateDevice):
    def __init__(self, feature):
        super().__init__(feature)
        ClimateDevice.__init__(self)
        pass

    """Representation of a BleBox climate feature."""

    @property
    def supported_features(self):
        """Return the supported climate features."""
        return SUPPORT_TARGET_TEMPERATURE

    @property
    def hvac_mode(self):
        """Return the desired HVAC mode."""
        return {None: None, False: HVAC_MODE_OFF, True: HVAC_MODE_HEAT}[
            self._feature.is_on
        ]

    @property
    def hvac_action(self):
        """Return the actual current HVAC action."""
        on = self._feature.is_on
        if not on:
            return None if on is None else CURRENT_HVAC_OFF

        states = {None: None, False: CURRENT_HVAC_IDLE, True: CURRENT_HVAC_HEAT}

        heating = self._feature.is_heating
        return states[heating]

    @property
    def hvac_modes(self):
        """Return a list of possible HVAC modes."""
        return (HVAC_MODE_OFF, HVAC_MODE_HEAT)

    @property
    def temperature_unit(self):
        """Return the temperature unit."""
        return TEMP_CELSIUS

    @property
    def current_temperature(self):
        """Return the current temperature."""
        return self._feature.current

    @property
    def max_temp(self):
        """Return the maximum thermostat setting."""
        return self._feature.max_temp

    @property
    def min_temp(self):
        """Return the minimum thermostat setting."""
        return self._feature.min_temp

    @property
    def target_temperature(self):
        """Return the desired thermostat temperature."""
        return self._feature.desired

    async def async_set_hvac_mode(self, hvac_mode):
        """Set the climate entity mode."""
        modemap = {HVAC_MODE_OFF: "async_off", HVAC_MODE_HEAT: "async_on"}
        await getattr(self._feature, modemap[hvac_mode])()

    async def async_set_temperature(self, **kwargs):
        """Set the thermostat temperature."""
        value = kwargs[ATTR_TEMPERATURE]
        await self._feature.async_set_temperature(value)


class TestSauna(DefaultBoxTest):
    """Tests for entities representing a BleBox saunaBox."""

    DEVCLASS = "climates"
    ENTITY_CLASS = BleBoxClimateEntity

    DEV_INFO_PATH = "api/heat/extended/state"
    DEVICE_EXTENDED_INFO_PATH = "/state/extended"

    DEVICE_INFO = json.loads(
        """
    {
        "device": {
            "deviceName": "My SaunaBox",
            "type": "saunaBox",
            "fv": "0.176",
            "hv": "0.6",
            "apiLevel": "20180604",
            "id": "1afe34db9437",
            "ip": "192.168.1.11"
        }
    }
    """
    )
    DEVICE_INFO_THERMO = json.loads(
        """
        {
          "device": {
            "deviceName": "My ThermoBox",
            "type": "thermoBox",
            "product": "thermoBox",
            "hv": "thB.1.0",
            "fv": "0.1031",
            "universe": 0,
            "apiLevel": "20200229",
            "iconSet": 43,
            "categories": [
              7
            ],
            "id": "f6cfa2f11cd3",
            "ip": "192.168.49.183",
            "availableFv": null
          }
        }
        """
    )
    DEVICE_EXTENDED_INFO_THERMO = json.loads(
        """
    {
      "thermo": {
        "state": 0,
        "operatingState": 3,
        "desiredTemp": -70,
        "mode": 2,
        "minimumTemp": -1230,
        "maximumTemp": 6000,
        "safety": {
          "eventReason": 0,
          "triggered": [

          ]
        },
        "safetyTempSensor": {
          "sensorId": 1
        }
      },
      "sensors": [
        {
          "id": 0,
          "type": "temperature",
          "value": 2098,
          "state": 2
        },
        {
          "id": 1,
          "type": "temperature",
          "value": 2775,
          "state": 2
        }
      ]
    }
    """
    )

    def patch_version(apiLevel):
        """Generate a patch for a JSON state fixture."""
        return f"""
        {{ "device": {{ "apiLevel": {apiLevel} }} }}
        """

    DEVICE_INFO_FUTURE = jmerge(DEVICE_INFO, patch_version(future_date()))
    DEVICE_INFO_LATEST = jmerge(
        DEVICE_INFO, patch_version(get_latest_api_level("saunaBox"))
    )
    DEVICE_INFO_UNSUPPORTED = jmerge(DEVICE_INFO, patch_version(20180603))

    DEVICE_INFO_UNSPECIFIED_API = json.loads(
        """
    {
        "device": {
            "deviceName": "My SaunaBox",
            "type": "saunaBox",
            "fv": "0.176",
            "hv": "0.6",
            "id": "1afe34db9437",
            "ip": "192.168.1.11"
        }
    }
    """
    )

    def patch_state(state, current, desired):
        """Generate a patch for a JSON state fixture."""
        return f"""
        {{
            "heat": {{
                "state": {state},
                "desiredTemp": {desired},
                "sensors": [ {{ "value": {current} }} ]
            }}
        }}
        """

    STATE_DEFAULT = json.loads(
        """
    {
        "heat": {
            "state": 0,
            "desiredTemp": 6428,
            "maximumTemp": 12166,
            "minimumTemp": -5166,
            "sensors": [
                {
                    "type": "temperature",
                    "id": 0,
                    "value": 3996,
                    "trend": 0,
                    "state": 2,
                    "elapsedTimeS": 0
                }
            ]
        }
    }
    """
    )

    STATE_OFF_BELOW = STATE_DEFAULT
    STATE_NEEDS_HEATING = jmerge(STATE_DEFAULT, patch_state(1, 2320, 3871))

    STATE_OFF_ABOVE = jmerge(STATE_DEFAULT, patch_state(0, 3871, 2876))
    STATE_NEEDS_COOLING = jmerge(STATE_DEFAULT, patch_state(1, 3871, 2876))

    STATE_REACHED = jmerge(STATE_DEFAULT, patch_state(1, 2320, 2320))
    STATE_THERMO_SET = jmerge(STATE_DEFAULT, patch_state(1, 2320, 4320))

    async def test_init(self, aioclient_mock):
        """Test default state."""

        await self.allow_get_info(aioclient_mock)
        entity = (await self.async_entities(aioclient_mock))[0]

        assert entity.name == "My SaunaBox (saunaBox#thermostat)"
        assert entity.unique_id == "BleBox-saunaBox-1afe34db9437-thermostat"

        assert entity.device_class is None
        assert entity.supported_features & SUPPORT_TARGET_TEMPERATURE
        assert entity.hvac_modes == (HVAC_MODE_OFF, HVAC_MODE_HEAT)

        assert entity.hvac_mode is None
        assert entity.hvac_action is None
        assert entity.target_temperature is None
        assert entity.temperature_unit == TEMP_CELSIUS
        assert entity.state is None
        assert entity.max_temp is None
        assert entity.min_temp is None

    async def test_thermo_init(self, aioclient_mock):
        """Test initialisation with device state"""
        self.DEVICE_INFO = self.DEVICE_INFO_THERMO
        self.DEVICE_EXTENDED_INFO = self.DEVICE_EXTENDED_INFO_THERMO
        await self.allow_get_info(
            aioclient_mock,
        )
        entity = (await self.async_entities(aioclient_mock))[0]
        assert entity.device_info["name"] == "My ThermoBox"

    async def test_device_info(self, aioclient_mock):
        await self.allow_get_info(aioclient_mock, self.DEVICE_INFO)
        entity = (await self.async_entities(aioclient_mock))[0]
        assert entity.device_info["name"] == "My SaunaBox"
        assert entity.device_info["mac"] == "1afe34db9437"
        assert entity.device_info["manufacturer"] == "BleBox"
        assert entity.device_info["model"] == "saunaBox"
        assert entity.device_info["sw_version"] == "0.176"

    async def test_update(self, aioclient_mock):
        """Test updating."""

        entity = await self.updated(aioclient_mock, self.STATE_DEFAULT)

        assert entity.hvac_mode == HVAC_MODE_OFF
        assert entity.hvac_action == CURRENT_HVAC_OFF
        assert entity.target_temperature == 64.3
        assert entity.current_temperature == 40.0
        assert entity.temperature_unit == TEMP_CELSIUS
        assert entity.max_temp == 121.7
        assert entity.min_temp == -51.7

    async def test_on_when_below_target(self, aioclient_mock):
        """Test when temperature is below desired."""

        entity = await self.updated(aioclient_mock, self.STATE_OFF_BELOW)
        assert entity.state == entity.hvac_mode == HVAC_MODE_OFF
        assert entity.hvac_action == CURRENT_HVAC_OFF

        self.allow_get(aioclient_mock, "/s/1", self.STATE_NEEDS_HEATING)
        await entity.async_set_hvac_mode(HVAC_MODE_HEAT)

        assert entity.target_temperature == 38.7
        assert entity.current_temperature == 23.2
        assert entity.state == entity.hvac_mode == HVAC_MODE_HEAT
        assert entity.hvac_action == CURRENT_HVAC_HEAT

    async def test_on_when_above_target(self, aioclient_mock):
        """Test when temperature is below desired."""

        entity = await self.updated(aioclient_mock, self.STATE_OFF_ABOVE)
        assert entity.state == entity.hvac_mode == HVAC_MODE_OFF
        assert entity.hvac_action == CURRENT_HVAC_OFF

        self.allow_get(aioclient_mock, "/s/1", self.STATE_NEEDS_COOLING)
        await entity.async_set_hvac_mode(HVAC_MODE_HEAT)

        assert entity.target_temperature == 28.8
        assert entity.current_temperature == 38.7
        assert entity.state == entity.hvac_mode == HVAC_MODE_HEAT
        assert entity.hvac_action == CURRENT_HVAC_IDLE

    async def test_on_when_at_target(self, aioclient_mock):
        """Test when temperature is below desired."""

        entity = await self.updated(aioclient_mock, self.STATE_OFF_ABOVE)
        assert entity.state == entity.hvac_mode == HVAC_MODE_OFF
        assert entity.hvac_action == CURRENT_HVAC_OFF

        self.allow_get(aioclient_mock, "/s/1", self.STATE_REACHED)
        await entity.async_set_hvac_mode(HVAC_MODE_HEAT)

        assert entity.target_temperature == 23.2
        assert entity.current_temperature == 23.2
        assert entity.state == entity.hvac_mode == HVAC_MODE_HEAT
        assert entity.hvac_action == CURRENT_HVAC_IDLE

    async def test_off(self, aioclient_mock):
        """Test turning off."""

        entity = await self.updated(aioclient_mock, self.STATE_REACHED)

        self.allow_get(aioclient_mock, "/s/0", self.STATE_OFF_BELOW)
        await entity.async_set_hvac_mode(HVAC_MODE_OFF)

        assert entity.target_temperature == 64.3
        assert entity.current_temperature == 40.0
        assert entity.state == entity.hvac_mode == HVAC_MODE_OFF
        assert entity.hvac_action == CURRENT_HVAC_OFF

    async def test_set_thermo(self, aioclient_mock):
        """Test setting thermostat."""

        entity = await self.updated(aioclient_mock, self.STATE_REACHED)
        self.allow_get(aioclient_mock, "/s/t/4321", self.STATE_THERMO_SET)
        await entity.async_set_temperature(**{ATTR_TEMPERATURE: 43.21})
        assert entity.current_temperature == 23.2  # no change yet
        assert entity.target_temperature == 43.2
        assert entity.state == entity.hvac_mode == HVAC_MODE_HEAT
        assert entity.hvac_action == CURRENT_HVAC_HEAT



from .sensor import Temperature
from typing import Optional, Any, Union
from .feature import Feature
from blebox_uniapi.jfollow import follow


class Climate(Temperature):
    _is_on: Optional[bool]
    _desired: Union[float, int, None]
    _is_heating: Optional[bool]
    _is_cooling: Optional[bool]
    _min_temp: Union[float, int, None]
    _max_temp: Union[float, int, None]
    _mode: Optional[int]
    _havc_action: Optional[int]

    def __init__(self, product, alias, methods, mode):
        super().__init__(product, alias, methods)
        self._mode = mode

    @property
    def mode(self) -> Optional[int]:
        return self._mode

    @property
    def is_on(self) -> Optional[bool]:
        return self._is_on

    @property
    def desired(self) -> Any:
        return self._desired

    @property
    def current(self) -> Any:
        return self._current

    @property
    def max_temp(self) -> Union[float, int, None]:
        return self._max_temp

    @property
    def min_temp(self) -> Union[float, int, None]:
        return self._min_temp

    @property
    def is_heating(self) -> Optional[bool]:
        return self._is_heating

    @property
    def is_cooling(self) -> Optional[bool]:
        return self._is_cooling

    @property
    def hvac_action(self) -> Optional[int]:
        return self._havc_action

    @classmethod
    def many_from_config(
        cls, product, box_type_config, extended_state
    ) -> list["Feature"]:
        # note: by default single config entry yields single feature instance but certain feature
        # domains (e.g. lights) may handle this differently depending on their `extended_state`
        alias, methods = box_type_config[0]
        if extended_state is not None:
            safetyIdPath = box_type_config[0][1].get("safetySensorId")
            if safetyIdPath:
                safety_sensor_id = follow(extended_state, safetyIdPath)
                temp_sensor_id = cls.get_temp_sensor_id(
                    safety_sensor_id, extended_state["sensors"]
                )
                methods = Feature.resolve_access_method_paths(methods, temp_sensor_id)

            args = [alias, methods]
            mode = extended_state.get("thermo", {}).get("mode", 1)
            return [cls(product, *args, mode)]
        return [cls(product, *args, 1) for args in box_type_config]

    async def async_on(self) -> None:
        await self.async_api_command("on")

    async def async_off(self) -> None:
        await self.async_api_command("off")

    async def async_set_temperature(self, value: Any) -> None:
        await self.async_api_command("set", int(round(value * 100.0)))

    def _read_is_on(self) -> Optional[bool]:
        product = self._product
        if product.last_data is not None:
            raw = self.raw_value("state")
            if raw is not None:  # no reading
                alias = self._alias
                return product.expect_int(alias, raw, 3, 0) in (
                    1,
                    3,
                )  # 1: On, 3: Boost(thermoBox max temp of mode)
        return None

    def _read_operating_state(self) -> Optional[int]:
        """Return current operating state"""
        if self._product.last_data is not None:
            try:
                raw = self.raw_value("operatingState")
                if raw in range(0, 4) and isinstance(raw, int):
                    return raw
            except KeyError:
                return None

    def _read_is_heating(self) -> Optional[bool]:
        if not self._product.last_data:
            return None

        return self.is_on and (self.current < self.desired)

    def _read_is_cooling(self) -> Optional[bool]:
        if not self._product.last_data:
            return None

        return self.is_on and (self.current > self.desired)

    def _read_mode(self) -> Optional[int]:
        if self._product.last_data is not None:
            try:
                raw = self.raw_value("mode")
                if raw in range(0, 3) and isinstance(raw, int):
                    return raw
            except KeyError:
                return 1
        return None

    def after_update(self) -> None:
        self._is_on = self._read_is_on()
        self._desired = self._read_temperature("desired")
        self._current = self._read_temperature("temperature")
        self._is_heating = self._read_is_heating()
        self._is_cooling = self._read_is_cooling()
        self._havc_action = self._read_operating_state()
        if self._product.last_data is None:
            self._min_temp = None
            self._max_temp = None
            return

        raw_min = self.raw_value("minimum")
        if raw_min is None:
            return

        self._min_temp = self._read_temperature("minimum")
        self._max_temp = self._read_temperature("maximum")

    @staticmethod
    def get_temp_sensor_id(safety_sensor_id: int, sensor_list) -> Optional[int]:
        """Return ID of the first sensor which is not a safety sensor."""
        li_sensor_id = [
            sensor.get("id")
            for sensor in sensor_list
            if sensor.get("id") is not None and sensor.get("id") != safety_sensor_id
        ]
        li_sensor_id.sort()

        if not li_sensor_id:
            return None

        return li_sensor_id[0]



"""Tests for the Update feature."""

import pytest
from unittest.mock import AsyncMock, MagicMock

from blebox_uniapi.update import Update
from blebox_uniapi import error


@pytest.fixture
def mock_product():
    product = MagicMock()
    product.firmware_version = "0.176"
    product.available_firmware_version = None
    return product


@pytest.fixture
def update(mock_product):
    return Update(mock_product, "update", {})


class TestInstalledVersion:
    def test_returns_firmware_version(self, update, mock_product):
        assert update.installed_version == "0.176"

    def test_returns_none_when_not_set(self, update, mock_product):
        mock_product.firmware_version = None
        assert update.installed_version is None


class TestLatestVersion:
    def test_returns_none_when_not_set(self, update, mock_product):
        assert update.latest_version is None

    def test_returns_available_firmware_version(self, update, mock_product):
        mock_product.available_firmware_version = "1.2.3"
        assert update.latest_version == "1.2.3"


class TestAsyncUpdate:
    async def test_calls_ota_check_on_product(self, update, mock_product):
        mock_product.async_ota_check = AsyncMock()
        await update.async_update()
        mock_product.async_ota_check.assert_called_once()

    async def test_propagates_connection_error_from_ota_check(self, update, mock_product):
        mock_product.async_ota_check = AsyncMock(
            side_effect=error.ConnectionError("connection refused")
        )
        with pytest.raises(error.ConnectionError):
            await update.async_update()


class TestAsyncInstall:
    async def test_calls_ota_update_on_product(self, update, mock_product):
        mock_product.async_ota_update = AsyncMock()
        await update.async_install()
        mock_product.async_ota_update.assert_called_once()

    async def test_propagates_connection_error_from_ota_update(self, update, mock_product):
        mock_product.async_ota_update = AsyncMock(
            side_effect=error.ConnectionError("connection refused")
        )
        with pytest.raises(error.ConnectionError):
            await update.async_install()
